package org.eso.ohs.core.dbb.client;

// $Id: Coordinate.java,v 1.3 2005/04/11 13:53:44 tcanavan Exp $

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import org.eso.ohs.core.utilities.TextUtils;

/**
 *  A transducer to convert set/get of String values into set/get
 *  on an int property.  The target property's name is given to the
 *  constructor.
 *
 *  <p>
 *  For example, users will call it's
 *  setValue/getValue methods.  Those calls will be translated into
 *  calls to setX/getX on the object given to the constructor.
 *
 *  <p>
 */

public class Coordinate implements DbbDataType {

    /** Default zero coordinate */
    private final static String DEFAULT_ANGLE_ = "-00:00:00.000";

    /** Milliseconds per second: a constant to improve readability */
    final int MILLISECS_PER_SEC = 1000;
    
    /** Milliseconds per minute: a constant to improve readability */
    final int MILLISECS_PER_MINUTE = 60 * MILLISECS_PER_SEC;
    
    /** Milliseconds per unit (hour or degree) */
    final int MILLISECS_PER_UNIT = 60 * MILLISECS_PER_MINUTE;
    
    /** Number of decimal places of accuracy */
    final int DECIMAL_PLACES_ACCURACY = 3;

    //----------------------------------------
    // Public methods
    //----------------------------------------

    /** Check if the given value is correct */
    public void checkBounds(int sign, int hours, int minutes,
			    int seconds, int milliSecs) {

	if (sign < 0 || hours >= 24 || minutes >= 60
	    || seconds >= 60 || milliSecs >= 1000) {
	    throw new IllegalArgumentException("Illegal time value");
	}
    }


    /**
     * Converts data to a String from the meta-level equivalent of the
     * actual property's type, i.e. from a java.lang.Long
     * to a long int.  Used with reflection to get
     * the current property value.
     *
     * @param value	value of the property.
     */
    public String fromDatabaseData(Object obj) {
	if (obj == null) {
	    return "00:00:00.0000";
	}
	long units, minutes;
	long milliSeconds = ((Integer)obj).intValue();
	boolean negative = false;

	if (milliSeconds < 0) {
	    negative = true;
	    milliSeconds = -milliSeconds;
	}

	units = milliSeconds/MILLISECS_PER_UNIT;
	milliSeconds -= units * MILLISECS_PER_UNIT;

	minutes = milliSeconds/MILLISECS_PER_MINUTE;
	milliSeconds -= minutes * MILLISECS_PER_MINUTE;

	return (negative ? "-" : " ") +
	    TextUtils.zeroPadded(2, units) + ":" +
	    TextUtils.zeroPadded(2, minutes) + ":" +
	    TextUtils.zeroPadded(2, (milliSeconds/MILLISECS_PER_SEC)) + "." +
	    TextUtils.zeroPadded(3, (milliSeconds%MILLISECS_PER_SEC));
    }


    /**
     * Converts data from a String to a meta-level equivalent of the
     * actual property's type, i.e. to a java.lang.Integer
     * representing an int.  Used with reflection to set
     * the new property value.
     *
     * @param value the new value of the property.
     */
    public Object toDatabaseData(String value) {
	int result = 0;
	StringTokenizer st = new StringTokenizer(value, " :");

	try {
	    int units;
	    String unitsString = st.nextToken();
	    int minutes = 0;
	    int seconds = 0;
	    int milliSecs = 0;
	    int sign = +1;

	    if (st.hasMoreTokens()) {
		minutes = new Integer(st.nextToken()).intValue();
		if (st.hasMoreTokens()) {
		    seconds = new Integer(st.nextToken(": ,.")).intValue();
		    if (st.hasMoreTokens()) {
			String ms = st.nextToken(".,");
			for (int j=1; j <= DECIMAL_PLACES_ACCURACY; j++) 
			    ms += "0";
			ms = ms.substring(0, DECIMAL_PLACES_ACCURACY);
			milliSecs = new Integer(ms).intValue();
		    }
		}
	    }

	    if (unitsString.startsWith("-")) {
		sign = -1;
		units = -new Integer(unitsString).intValue();
	    }
	    else {
		units = new Integer(unitsString).intValue();
	    }

	    checkBounds(sign, units, minutes, seconds, milliSecs);

	    result = sign * (units * MILLISECS_PER_UNIT + 
			     minutes * MILLISECS_PER_MINUTE + 
			     seconds * MILLISECS_PER_SEC + milliSecs);
	}
	catch (NumberFormatException e) {
	    throw new IllegalArgumentException(e.getMessage());
	}
	catch (NoSuchElementException e) {
	    throw new IllegalArgumentException(e.getMessage());
	}

	return new Integer(result);
    }


    /** @return the display data type */
    public Class getDataType() {
	return String.class;
    }


    /** @return the database data type */
    public Class getDatabaseType() {
	return Integer.class;
    }

    /** @return the display data type length */
    public int getDataTypeLength() {
	return DEFAULT_ANGLE_.length();
    }
}

