package org.eso.ohs.core.dbb.server;

// $Id: DbbSession.java,v 1.4 2005/06/16 15:54:16 gdonino Exp $

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbSqlTable;
import org.eso.ohs.core.utilities.ConnectionPool;

/**
 * This class allows to set up database connections on more than one
 * database, to retrieve and cash database metadata for a given
 * database, and to run sql queries..
 *
 * @author Carlo Boarotto, ESO 14-Sep-1999
 * 
 **/
public class DbbSession {
private static final String rcsid = "$Id: DbbSession.java,v 1.4 2005/06/16 15:54:16 gdonino Exp $";
private static Logger stdlog_ = 
    Logger.getLogger( DbbSession.class );
 
 
    /** JDBC URL - database system */
    private String url_ = null;

    /**
     * Hashtable containing the database name as key and the related
     * connection as element.
     */
    private Hashtable connHashtable_ = new Hashtable();

    /**
     * Hashtable containing the database name as key and the related
     * metadata as element. 
     */
    private Hashtable dbmdHashtable_ = new Hashtable();

    /**
     * Default database name to be used when no database name is
     * specified.
     */
    private String database_;

    /**
     * Default database owner to be used when no database owner is
     * specified. 
     */
    private String owner_ = null;

    /** Connection user name */
    private String username_ = null;

    /** Connection password */
    private String password_ = null;
   
    /** Current connection */
    private Connection conn_ = null;


    //----------------------------------------
    // Constructor
    //----------------------------------------

    /**
     * Build a DbbSession object setting its internal variables and the
     * database drivers.
     *
     * @param url		JDBC URL.
     * @param default_db	Default database.
     * @param username	Login name.
     * @param password	Login password.
     */
    public DbbSession(String url,
		      String default_db,
		      String username,
		      String password) {        
	url_ = url;
	username_ = username;
	password_ = password;
	database_ = default_db;
	// database_ = new String(default_db);
	// stdlog_.debug("..........Database " + database_);
    }


    //----------------------------------------
    // Public methods
    //----------------------------------------

    /** @return database_ */
    public String getDefaultDatabase() { return database_; }


    /** @return owner_ */
    public String getDefaultOwner() { return owner_; }


    /** @return conn_ */
    public Connection getCurrentConnection() {return conn_;}


    /** Release the current connection */
    public void releaseCurrentConnection() {
	if(conn_ != null)
	    ConnectionPool.getPool().releaseConnection(conn_);
	conn_ = null;
    }

    /**
     * Define which database account should be used by this object
     * for all future operations. 
     * @param username  Login name for the user.
     * @param password  Password for that login name.
     */
    // AMC, 10-Sep-2003: added
    public void setAccount( String username, String password ) {
        username_ = username;
        password_ = password;
        releaseCurrentConnection();
    }
    
    /**
     * Get database metadata information for a given database. If the
     * metadata have been already stored in the hashtable return them
     * directly, otherwise retrieve them from the database and save in
     * the hashtable.
     *
     * @param db_name	Database name.
     *
     * @return A DatabaseMetaData object.  
     * @exception SQLException JDBC exception
     */
    public DatabaseMetaData getDatabaseMetadata(String db_name) 
	throws SQLException { 

	/* Switch to the default database if db_name is either empty or
	 * null 
	 */
	if(db_name == null || db_name.equals(""))
	    db_name = database_;

	DatabaseMetaData dbmd =
	    (DatabaseMetaData)dbmdHashtable_.get(db_name);

	if (dbmd == null) {
	    //try {


	    conn_ = ConnectionPool.getPool().
		acquireConnection(url_,username_,password_);
    
	    dbmd = conn_.getMetaData();
 
	    dbmdHashtable_.put(db_name, dbmd);
	    ConnectionPool.getPool().releaseConnection(conn_);
	    conn_ = null;

	}
	return dbmd;
    }


    /**
     * Get a the metadata from the default database.
     *
     * @return A DatabaseMetaData object.
     * @exception SQLException JDBC exception
     */
    public DatabaseMetaData getDatabaseMetadata() throws SQLException {
	return getDatabaseMetadata(database_);
    }


    /**
     * Get the java class that should be associated with a specific
     * column from the database metadata. 
     * 
     * @param table   Database table object.
     * @param column  Table column name.
     *
     * @return Java class that maps the database column type.
     * @exception SQLException JDBC exception
     */
    public Class getColumnDataType(DbbSqlTable table,
				   String column) throws SQLException {

	int type = -1;
	ResultSet rs = null;

	table.completeTable(this);
	DatabaseMetaData dbmd = getDatabaseMetadata(table.getDatabase());

	//   try {

	rs = dbmd.getColumns(table.getDatabase(), 
			     table.getOwner(), 
			     table.getTable(), 
			     column);

	while (rs.next())
	    type = rs.getInt(5);
	//    }
	//    catch(SQLException e) {
	//     stdlog_.warn("",e);
	//    }

	Class c = getDataTypeClass(type);

	//     stdlog_.debug("Table: " + table + " Column: " +
	// 		       column + " Class: " + c);
	return c;
    }


    /**
     * Run a sql query and return the result.
     *
     * @param db_name		Database where to execute the sql statement.
     * @param sql			String containing the sql statement.
     * @param connect_again	Boolean flag: if true force to 
     *				connect again.
     *
     * @return A ResultSet object.  
     * @exception SQLException JDBC exception
     */
    public ResultSet execSql(String db_name, String sql, boolean connect_again)
	throws SQLException {

	Connection con = null;
	Statement stmt = null;
	ResultSet rs = null;
    SQLException sqlExec = null;
	conn_ = ConnectionPool.getPool().acquireConnection(
							   url_,
							   username_,
							   password_
							   );
	stmt = conn_.createStatement();
    try {
        rs = stmt.executeQuery(sql);
    }
    catch(SQLException e) {
        e.printStackTrace();
        sqlExec = e;      
    }
    
	//rs = stmt.executeQuery(sql);
  
       ConnectionPool.getPool().releaseConnection(conn_);
	   conn_ = null;
       if(sqlExec != null)
           throw new SQLException(sqlExec.getMessage());
	   return rs;
    }


    /**
     * Run a sql query and return the result.
     *
     * @param db_name	Database where to execute the sql statement.
     * @param sql		String containing the sql statement.
     *
     * @return A ResultSet object.  
     * @exception SQLException JDBC exception
     */
    public ResultSet execSql(String db_name, String sql)
	throws SQLException {
	return execSql(db_name, sql, false);
    }


    /**
     * Run a sql query on the default database and return the result.
     *
     * @param sql			String containing the sql statement.
     * @param connect_again	Boolean flag: if true force to 
     *				connect again.
     *
     * @return A ResultSet object.  
     * @exception SQLException JDBC exception
     */
    public ResultSet execSql(String sql, boolean connect_again)
	throws SQLException {
	return execSql(database_, sql, connect_again);
    }


    /**
     * Run a sql query on the default database and return the result.
     *
     * @param sql			String containing the sql statement.
     *
     * @return A ResultSet object.
     * @exception SQLException JDBC exception
     */
    public ResultSet execSql(String sql) throws SQLException {
	return execSql(database_, sql, false);
    }

    public void execUpdateSql(String sql) throws SQLException {
	Connection con = null;
	Statement stmt = null;

	conn_ = ConnectionPool.getPool().acquireConnection(url_,
							   username_,
							   password_);
	stmt = conn_.createStatement();
	stmt.executeUpdate(sql);
	ConnectionPool.getPool().releaseConnection(conn_);
	conn_ = null;
    }
	
	 public void execSqlNoResulset(String sql) throws SQLException {
		Connection con = null;
		Statement stmt = null;

		conn_ = ConnectionPool.getPool().acquireConnection(url_,
									   username_,
									   password_);
		stmt = conn_.createStatement();
		stmt.execute(sql);
		ConnectionPool.getPool().releaseConnection(conn_);
		conn_ = null;
		    }

    /**
     * Utility to map database types to java classes.
     *
     * @param type Database type (as specified in java.sql.Types).
     * @return The corresponding java class.
     */
    public static Class getDataTypeClass(int type) {

	switch(type) {

	case Types.CHAR:
	case Types.VARCHAR:
	case Types.LONGVARCHAR:
	    return String.class;

	case Types.NUMERIC:
	case Types.DECIMAL:
	    return Long.class;
	    // return BigDecimal.class;
      
	case Types.BIT:
	    return Boolean.class;
      
	case Types.TINYINT:
	case Types.SMALLINT:
	case Types.INTEGER:
	    return Integer.class;
      
	case Types.BIGINT:
	    return Long.class;
      
	case Types.REAL:
	    return Float.class;
      
	case Types.DOUBLE:
	    return Double.class;
      
	case Types.BINARY:
	case Types.VARBINARY:
	case Types.LONGVARBINARY:
	    return Object.class;
	}

	return Object.class;
    }


    //----------------------------------------
    // Debug methods
    //----------------------------------------

    /**
     * Debug: print out a ResultSet object.
     */
    public static void dumpRs(String header, ResultSet rs) {

	try {

	    ResultSetMetaData rsmd = rs.getMetaData();
	    int numberOfColumns = rsmd.getColumnCount();

	    stdlog_.debug(header);
	    StringBuffer printOut = new StringBuffer();
	    while (rs.next()) {
		for (int i = 1; i <= numberOfColumns; i++) {
		    printOut.append("   ");
		    printOut.append(rs.getString(i));
		    printOut.append("  ");
		}
	    }
	    stdlog_.debug(printOut.toString());
	}
	catch (SQLException e) {
	    stdlog_.warn("",e);
	}
    }

    /**
     * Debug: get an array with all the table names.
     * @exception SQLException JDBC exception
     */
    public String[] getTables(String db_name)
	throws SQLException {

	    ResultSet rs;
	    String[] types = {"TABLE"};
	    String[] info_array = null;
	    int len = 0;

	    // Temporary vector to store data
	    Vector info_vector = new Vector();

	    rs = getDatabaseMetadata(db_name).getTables(null, 
							null, 
							null, 
							types);

	    while (rs.next()) {
		String s = rs.getString(3);
		info_vector.addElement(s);
	    }

	    rs.close();
	    releaseCurrentConnection();

	    // Dump the vector into the return array
	    len = info_vector.size();
	    info_array = new String[len];
	    info_vector.copyInto(info_array);

	    return info_array;
	}

    /**
     * Debug main to run regression tests
     */
    public static void main(String[] args) {
    }
}
