package org.eso.ohs.instruments;

import org.eso.ohs.core.utilities.Convert;

/**
 * This class stores Coordinate parameters for RA and DEC
 * as strings. RA is stored ad and DEC is stored as
 * <p>
 * $Id: CoordParameter.java,v 1.21 2004/02/17 14:31:57 tcanavan Exp $
 * @author Tim Canavan
 */

public class CoordParameter extends Parameter {

    static final long serialVersionUID = -3984228626336657848L;

    /** Construct a coordinate class */
    public CoordParameter(ParamMetaData pMeta) {
        super(pMeta);
    }

    /**
     * Return tooltip text to be displayed in the OD
     * and OB grids for a given paraeter
     *
     * @return Tooltip for coord parameter
     */
    public String getToolTipText() {
        String range = getRange();
        if(range == null)
            return "The TSF contains a wrong PARAMETER definition";
        if(range.equalsIgnoreCase("ra"))
            return "Enter Right Ascension value";
        if(range.equalsIgnoreCase("dec"))
            return "Enter Declination value";
        return "The TSF contains a wrong PARAMETER definition";
    }

   /**
    * Overrides the original getRange() from class Parameter
    * evaluates the ParameterName and returnes "dec" or "ra"
    * accordingly. In case of error it returns null. 
    *
    * <br>ChangeLog: Toni Schlichting 26-Jan-2001
    *
    * @return range parameters "ra" or "dec" which are meaningful return
    *         values or (String)null which indicates an error condition
    */

    public String getRange() {
       /*
        * The return values are because the java compiler does not permit to
        * have a case where the method returns nothing in contrast to its
        * declaration.  No exception is thrown, because the verify method,
        * which calls getRange() also returns strings, which propably is
        * more efficient than throwing an exception even though this would
        * be common java style.
        */
        boolean postfix = false;
        String parameterName = null;
        String ra       = "ra",
               dec      = "dec";
        parameterName = super.getParameterName();

        if(parameterName.endsWith("ALPHA"))
            return ra;
        if(parameterName.endsWith("DELTA"))
            return dec;
        return(String)null;
    }

    public char getDBTypeCode() {
        return 'C';
    }


    /**
     * This method verifies that a string is a valid Coordinate
     *
     * <br>ChangeLog: Toni Schlichting
     * <br>ChangeLog: cmavros 05/03/02, Change local converting to 
     *        call to Convert methods in Utilities
     *
     * @param value String repesenting a particular Coordinate
     * @return  boolean if coord is vailid
     */
    public boolean verifyValue(String value) {
        int firstValue = 0;
        String range = getRange();
        if(range == null)
            return false;

        if(value == null || value.equals("NODEFAULT")) {
            return false;
        }
        if(value.equals(getDefaultValue()) || value.length() == 0){
            return true;
        }

        try {
	    if(getRange().equalsIgnoreCase("ra")) {
            try {
		        firstValue = Convert.DDMMSSToMilliarcsec(value, true);
            }
            catch(IllegalArgumentException e) {
                return false;
            }
		    if(!Convert.isValidRA(firstValue)){
		        return false;
		    }

	    } else if (getRange().equalsIgnoreCase("dec")) {
            try {        
		        firstValue = Convert.DDMMSSToMilliarcsec(value, true);
            }
            catch(IllegalArgumentException e) {
                return false;
            }
		    if(!Convert.isValidDec(firstValue)){
		        return false;
		    }
	    } else
		    return false;

	}
	catch(NumberFormatException e) {
	    e.printStackTrace();
	    return false;
	}                             
        return true;
    }
}
