package org.eso.ohs.persistence;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.LinkedList;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.dfs.ObjectFactory;
import org.eso.ohs.dfs.Target;

/**
 * This class imports a target list from a text file 
 * and creates a set of observations blocks containing these targets 
 * 
 * @author pnunes 02.11.2005
 *
 */
public class ImportTargets {

		
	// regular expressions for the target parameters
	public static final String SEPARATOR = "\t";
	
	private static final String RA_HOURS = "((0*[0-1]?[0-9]|0*2[0-3])";
	private static final String DEC_DEGREES = "((\\+|-)?(0*[0-8]?[0-9]|0*90)";
	private static final String MINUTES ="(0*[0-5]?[0-9])";
	private static final String SECONDS ="((0*[0-5]?[0-9])(\\.[0-9]*)?))";
	
	public static final String RA_EXPR = 
		        RA_HOURS + ":" + MINUTES + ":" + SECONDS;
	
	public static final String RA_EXPR2 = 
        RA_HOURS + " " + MINUTES + " " + SECONDS;
	
	public static final String RA_EXPR3 = 
        RA_HOURS + "h " + MINUTES + "m " + SECONDS +"s";
	
	public static final String DEC_EXPR = 
		DEC_DEGREES + ":" + MINUTES + ":" + SECONDS;
	
	public static final String DEC_EXPR2 = 
		DEC_DEGREES + " " + MINUTES + " " + SECONDS;
	
	public static final String DEC_EXPR3 = 
		DEC_DEGREES + "d " + MINUTES + "m " + SECONDS +"s";;
	
	public static final String EQUINOX_EXPR = "((J|B)?([1-9]([0-9]{3})))?";
	public static final String NAME_EXPR = "(([a-zA-Z0-9(\\+)*(\\-)*( *)(\\.)*])*)";	
	public static final String PROPER_MOTION_EXPR = "(\\+|-)?([0-9]+)(\\.[0-9]*)?";
	public static final String EPOCH_EXPR = "([0-9]{4}(\\.[0-9]?))?";
	
	
	//error messages
	public static final String INVALID_PROPER_MOTION_FORMAT_ERROR_MSG = "Invalid proper motion format";
	public static final String INVALID_RA_FORMAT_ERROR_MSG = "Invalid RA format";
	public static final String INVALID_TARGET_NAME_ERROR_MSG = "Invalid Target name";
	public static final String INVALID_DEC_FORMAT_ERROR_MSG = "Invalid DEC format";
	public static final String INVALID_LINE_FORMAT_ERROR_MSG = "Invalid line format";
	public static final String INVALID_EQUINOX_FORMAT_ERROR_MSG = "Invalid equinox format";
	
	
	// default values 	
	public static final String DEFAULT_EQUINOX = "J2000";
	public static final String DEFAULT_EPOCH = "2000";
	
	/**
	 * Imports a target list file 
	 * 
	 * @param inputFile
	 * @return Target array
	 * @throws IOException
	 */
	public Target [] importTargets(File inputFile) throws IOException , TargetImportException{		
		Target [] res = null;		
		LinkedList targetList = new LinkedList();
		
		LineNumberReader reader = new LineNumberReader(new FileReader(inputFile)); 
				
		
		String line = reader.readLine();
		int lineNumber = reader.getLineNumber();
		
		
		while (line != null) {
			
			try {
				String [] values = parseLine(line);
				
				if (values.length > 0) {
					Target target = createTarget(values);
					targetList.add(target);
				}
				
				line = reader.readLine();
				lineNumber = reader.getLineNumber();
									
			} catch ( TargetImportException e) {
				throw new TargetImportException("Line " + lineNumber + ": " +  e.getMessage());
			}
		}
				
		res =  (Target [])targetList.toArray(new Target[targetList.size()]);
		return(res);				
	}
	
	
	/**
	 * Parses a line
	 * 
	 * Assumes the following sequence, with tab separators
	 *  Target name (mandatory), RA (mandatory), DEC (mandatory) , equinox (optional) , propRA (optional), propDEC (optional) , Epoch (optional)
	 *  
	 *   Other assumptions:
	 *    - if equinox is not there, it uses the default value 
	 * 	  - propDEC and propRA should always come together
	 *    - epoch is only to be there if there the two previous were defined, it also has a default when missing
	 * 
	 *   
	 * @param line
	 * @return a string array containing the values
	 * @throws TargetImportException
	 */
	protected String [] parseLine(String line) throws TargetImportException {
		String [] values;
		LinkedList valueList = new LinkedList();
		
		StringTokenizer tok = new StringTokenizer(line,SEPARATOR);
		Pattern [] pattern;
		
		if (tok.countTokens() == 0)
			return( new String[0]);
		else
			if (tok.countTokens() >0  &&  tok.countTokens() <3) {
				throw new TargetImportException(INVALID_LINE_FORMAT_ERROR_MSG);
			} else {
				
				// gets the target name - mandatory				
				if (  tok.countTokens() >=3 ) {
					String targetName = tok.nextToken();
					pattern = new Pattern[1];
					pattern[0] = Pattern.compile(NAME_EXPR);							
					matchPatterns(valueList,pattern,targetName,INVALID_TARGET_NAME_ERROR_MSG);					
					
					//	 gets the RA - mandatory, try the 3 different formats								
					String ra = tok.nextToken();
					pattern = getRAPatterns();
					matchPatterns(valueList,pattern,ra,INVALID_RA_FORMAT_ERROR_MSG);
					
					
					// gets the DEC - mandatory						
					String dec = tok.nextToken();
					pattern = getDECPatterns();								
					matchPatterns(valueList,pattern,dec,INVALID_DEC_FORMAT_ERROR_MSG);			
					
				}
				
				// check if next comes equinox, or something else
				if ( tok.hasMoreTokens() ) {
					
					boolean equinoxFound = false;				
					String token = tok.nextToken();
					
					// Equinox (optional) - if is not there use the default
					pattern = new Pattern[1];
					pattern[0] = Pattern.compile(EQUINOX_EXPR);
					
					//if matches add it to the list, otherwise use the default
					try {
						matchPatterns(valueList,pattern,token,INVALID_DEC_FORMAT_ERROR_MSG);
						equinoxFound = true;
					} catch (TargetImportException e) {
						valueList.add(DEFAULT_EQUINOX);
					}								
					
					
					// if the equinox was there, try to read whats next
					if (equinoxFound) {
						
						// propRA and propDEC are expected here
						if (tok.countTokens() >= 2) {						
							pattern = new Pattern[1];
							pattern[0] = Pattern.compile(PROPER_MOTION_EXPR);												
							String propRA = tok.nextToken(); 							
							matchPatterns(valueList,pattern,propRA,INVALID_PROPER_MOTION_FORMAT_ERROR_MSG);
							
						} else {
							throw new TargetImportException(INVALID_LINE_FORMAT_ERROR_MSG);	
						}	
						
						//it did not match equinox, it must be propRA
					} else {
						try {
							pattern = new Pattern[1];
							pattern[0] = Pattern.compile(PROPER_MOTION_EXPR);	
							String propRA = token;
						
							matchPatterns(valueList,pattern,propRA,INVALID_PROPER_MOTION_FORMAT_ERROR_MSG);
						} catch (TargetImportException e) {
							throw new TargetImportException(INVALID_LINE_FORMAT_ERROR_MSG);
						}
						
					}
					
					//	there must be propDEC next
					if (  tok.hasMoreTokens() ) {
						
						String propDEC = tok.nextToken();
						matchPatterns(valueList, pattern, propDEC,INVALID_PROPER_MOTION_FORMAT_ERROR_MSG); 
					} else {
						throw new TargetImportException(INVALID_LINE_FORMAT_ERROR_MSG);
					}
					
					
					// 	Epoch (optional) - if is not there use the default, 
					//  it is included only if propRA and propDEC where there
					if (  tok.hasMoreTokens() ) {
						pattern = new Pattern[1];
						pattern[0] = Pattern.compile(EPOCH_EXPR);
						String epoch = tok.nextToken();
						matchPatterns(valueList, pattern, epoch,"Invalid epoch format");
					} else {
						valueList.add(DEFAULT_EPOCH);
					}
					
					
					//not more tokens than RA and DEC, add default equinox
				} else {
					valueList.add(DEFAULT_EQUINOX);
					
				}
				
				
				
			}
		
		Object [] obj = valueList.toArray();
		values = new String[obj.length];
		for (int i = 0; i < values.length; i++) {
			values[i] = (String)obj[i];
		}		
		
		return(values);
		
	}


	/**
	 * Matches the value parameter against one of the patterns in the array
	 * If it doens't match any pattern, throws an exception
	 * Otherwise adds the value to the list 
	 * @param valueList
	 * @param patterns
	 * @param value
	 * @throws TargetImportException
	 */
	private void matchPatterns(LinkedList valueList, Pattern [] patterns, String value, String errorMessage) throws TargetImportException {
		boolean matched = false;
				
		for (int i = 0; i < patterns.length; i++) {
			Matcher matcher = patterns[i].matcher(value);	
			if (matcher.matches()) {
				matched = true;
				break;
			}
			
		}
		
		if (matched) {
			valueList.add(value);
		} else {
			throw new TargetImportException(errorMessage);
		}
	}
	
	/**
	 * Creates a Target object, assuming the following order on the array:
	 * 
	 * Target name, RA, DEC, Equinox, PropRA, PropDEC, Epoch
	 * the last three are optional
	 * 
	 * @param values
	 * @return
	 */
	private Target createTarget(String [] values ) {
		Target target = (Target) ObjectFactory.create(Target.class);
		
		String name = values[0];
		target.setName(name);
		
		
		int ra = parseRA(values[1]);		
		target.setRA(ra);
		
		int dec = parseDEC(values[2]);
		target.setDeclination(dec);
		
		String equinox = values[3];
		target.setEquinox(equinox);
		
		// if the array contains proper declinations
		if (values.length > 3) {
			
			float propRA = new Float(values[4]).floatValue();
			target.setPropRA(propRA);
			
			float propDEC = new Float(values[5]).floatValue();
			target.setProperDeclination(propDEC);
			
			float epoch = new Float(values[6]).floatValue();
			target.setEpoch(epoch);
		}
		 
		return(target);
	}
	
	
	/**
	 * @param ra
	 * @return
	 */
	protected int parseRA(String ra) {
		Pattern [] patterns = getRAPatterns();
		
		Matcher matcher = null;
		for (int i = 0; i < patterns.length; i++) {
			matcher = patterns[i].matcher(ra);	
			if (matcher.matches()) {
				break;
			}
		}
		
		String hours = matcher.group(2);
		String minutes = matcher.group(3);
		String seconds = matcher.group(4);		
		String s = hours +":" +  minutes +":" + seconds;
		int result = Convert.HHMMSSToMilliarcsec(s);
		
		return(result);
		
	}
	
	/**
	 * @param dec
	 * @return
	 */
	protected int parseDEC(String dec) {
		
		Pattern [] patterns = getDECPatterns();
		
		Matcher matcher = null;
		for (int i = 0; i < patterns.length; i++) {
			matcher = patterns[i].matcher(dec);	
			if (matcher.matches()) {
				break;
			}
		}
		
		String hours = null;
		if  (matcher.group(2) != null)
			hours = matcher.group(2) + matcher.group(3) ;
		else
			hours =matcher.group(3) ;
			
		String minutes = matcher.group(4);
		String seconds = matcher.group(5);
		
		String s = hours +":" + minutes +":" + seconds;
		int result = Convert.DDMMSSToMilliarcsec(s);
		
		return(result);

	}
	
	
	private Pattern [] getRAPatterns() {
		Pattern [] patterns = new Pattern[3];
		patterns = new Pattern[3];
		patterns[0] = Pattern.compile(RA_EXPR);
		patterns[1] = Pattern.compile(RA_EXPR2);
		patterns[2] = Pattern.compile(RA_EXPR3);
		return(patterns);
		
		
	}
	
	private Pattern [] getDECPatterns() {
		Pattern [] patterns = new Pattern[3];
		patterns = new Pattern[3];
		patterns[0] = Pattern.compile(DEC_EXPR);
		patterns[1] = Pattern.compile(DEC_EXPR2);
		patterns[2] = Pattern.compile(DEC_EXPR3);
		return(patterns);
		
		
	}
	
	
}
