package org.eso.ohs.persistence;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

import javax.swing.event.EventListenerList;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.ConstraintSet;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.dfs.Target;
import org.eso.ohs.persistence.StorageManager.MultiObjSpec;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;

/**
 * This class manages the in-memory copies of all StorableObjects.
 * When a reference to some StorableObject is requested by a client,
 * it first looks it up in a table to see if the object is already
 * in memory and, if so, returns it.  This ensures that only one
 * copy of any StorableObject is in memory at any one time, which will
 * eventually be written to the external media.  Thus an object cannot
 * appear to have several values at once.
 * 
 * <p>
 * If the request object is not in memory, it is retrieved from the
 * associated external media, added to the inMemory table and returned.
 * As part of the retrieval process, any StorableObjects referred to by
 * this object are also read into memory and referred to correctly.
 * 
 * <p>
 * For correct operation, there should be at most one instance of this
 * object.  Thus it is a singleton and can only be instantiated using
 * the static ObjectManager.getObjectManager() method.  Each time this
 * method is called it will return the <em>same</em> instance of the
 * ObjectManager.
 * 
 * <p>
 * $Id: ObjectManager.java,v 1.73 2005/01/14 13:57:00 tcanavan Exp $
 * <br>* @author Bob Kemp
 */
public class ObjectManager {

	/** Standard log channel */
	private static Logger stdlog_ = Logger.getLogger(ObjectManager.class);

	/** Invalid object id.  Used to indicate that there is no such object */
	public static final long NO_OBJECT = 0;

	/** Utility constant for invoking getter methods, using reflection */
	protected static final Object[] NO_ARGS = new Object[] {
	};

	/** A reference to the singleton ObjectManager instance. */
	protected static ObjectManager objMgr_ = new ObjectManager();

	protected Registry registry_;

	/**
	 * References to StorageManager objects for supported Media.
	 * The table is keyed by the Media value.  Supported media can be
	 * added at any time using the setStorageManager() method.
	 * 
	 * @see #setStorageManager
	 */
	protected Hashtable mediaMgr_ = new Hashtable();

	/* Support for handling object events */
	protected EventListenerList objectListeners;

	/* Support for handling directory events */
	protected EventListenerList dirListeners;

	/**
	 * Construct an instance of the ObjectManager.
	 * The constructor is protected, so that we can ensure there is
	 * only one instance of the ObjectManager, i.e. it's a singleton.
	 */
	protected ObjectManager() {
		objectListeners = new EventListenerList();
		dirListeners = new EventListenerList();

		createRegistry();
	}

	/**
	 * Return the one and only instance of the ObjectManager.
	 * 
	 * @return Singleton ObjectManager instance.
	 */
	public static ObjectManager getObjectManager() {
		return objMgr_;
	}

	/**
	 * Install a new instance of the ObjectManager.
	 * 
	 * @param objMgr	new ObjectManager instance.
	 */
	public static void setObjectManager(ObjectManager objMgr) {
		objMgr_ = objMgr;
	}

	/**
	 * Returns the appropiate Storage Manager for a particular Media type.
	 * 
	 * @param device The selected media that is being used
	 * 
	 * @return Storage manager for a particular device
	 */
	public StorageManager getStorageManager(Media device) {
		StorageManager stMgr = (StorageManager) mediaMgr_.get(device);

		if (stMgr == null) {
			throw new IllegalArgumentException(
				"unsupported media: " + device.toString());
		}
		return stMgr;
	}

	/**
	 * Set the StorageManager object for a given Media object.
	 * 
	 * @param device	the storage medium to be managed
	 * @param stMgr	the storage manager object
	 * 
	 * @exception ObjectIOException
	 * 			thrown if there is an I/O error while initialising
	 * 			the StorageManager
	 */
	public void setStorageManager(Media device, StorageManager stMgr)
		throws ObjectIOException {
		mediaMgr_.put(device, stMgr);
		stMgr.initialise();
	}

	/**
	 * Creates and installs a new instance of the Registry.
	 * Hook for subclasses.
	 */
	protected void createRegistry() {
		registry_ = new Registry();
	}

	/**
	 * Startup hook for subclasses.
	 * 
	 * @see #restart
	 * @see #shutdown
	 */
	public static void startup() {
		setObjectManager(new ObjectManager());
	}

	/**
	 * Shutdown hook for subclasses.
	 * 
	 * @see #startup
	 * @see #restart
	 */
	public void shutdown() {
		objMgr_ = null;
	}

	/**
	 * Re-creation hook for subclasses.
	 * 
	 * @see #startup
	 * @see #restart
	 */
	public void createNewInstance() {
		startup();
	}

	/**
	 * Shuts down the current ObjectManager and AutoSaver and then
	 * creates a new instance with the same StorageManager objects.
	 * 
	 * @exception ObjectIOException
	 *			if an I/O error occurred during the operation
	 * 
	 * @see #startup
	 * @see #shutdown
	 */
	public void restart() throws ObjectIOException {
		Media device;

		shutdown();
		createNewInstance();

		Enumeration enum = mediaMgr_.keys();

		while (enum.hasMoreElements()) {
			device = (Media) enum.nextElement();
			objMgr_.setStorageManager(
				device,
				(StorageManager) mediaMgr_.get(device));
		}
	}

	/**
	 * Returns a deep copy of the given tree of business objects
	 * after adding them to the Registry.
	 * 
	 * @param device	the storage medium to be managed
	 * @param dir	directory to contain the copy objects
	 * @param bo	top-level BusinessObject of 'tree' to copy
	 * 
	 * @exception ObjectIOException
	 * 			thrown if there is an I/O error while registering
	 * 
	 * @see BusinessObject#deepCopy
	 */
	public BusinessObject deepCopy(
		Media device,
		DirectoryNode dir,
		BusinessObject bo)
		throws ObjectIOException {
		return deepCopy(device, dir, bo, 0);
	}

	/**
	 * Returns a deep copy of the given tree of business objects
	 * after adding them to the Registry.
	 * 
	 * @param device	the storage medium to be managed
	 * @param dir	directory to contain the copy objects
	 * @param bo	top-level BusinessObject of 'tree' to copy
	 * @param baseId	Id on which to base the new object id,
	 * 			(StorageManager-dependent)
	 * 
	 * @exception ObjectIOException
	 * 			thrown if there is an I/O error while registering
	 * 
	 * @see BusinessObject#deepCopy
	 */
	public BusinessObject deepCopy(
		Media device,
		DirectoryNode dir,
		BusinessObject bo,
		long baseId)
		throws ObjectIOException {
		BusinessObject boCopy = bo.deepCopy();

		recursiveRegister(device, dir, boCopy, baseId);

		return boCopy;
	}

	/**
	 * Adds each element of the given tree of business objects
	 * to the Registry, if it's not already there.
	 * 
	 * @param device	the storage medium containing the objects
	 * @param dir	directory containing the objects
	 * @param bo	top-level BusinessObject of 'tree' to register
	 * 
	 * @exception ObjectIOException
	 * 			thrown if there is an I/O error while registering
	 * 
	 * @see #deepCopy
	 */
	public void recursiveRegister(
		Media device,
		DirectoryNode dir,
		BusinessObject bo)
		throws ObjectIOException {
		recursiveRegister(device, dir, bo, 0);
	}

	/**
	 * Adds each element of the given tree of business objects
	 * to the Registry, if it's not already there.
	 * 
	 * @param device	the storage medium containing the objects
	 * @param dir	directory containing the objects
	 * @param bo	top-level BusinessObject of 'tree' to register
	 * @param baseId	Id on which to base the new object id,
	 * 			(StorageManager-dependent)
	 * 
	 * @exception ObjectIOException
	 * 			thrown if there is an I/O error while registering
	 * 
	 * @see #deepCopy
	 */
	public void recursiveRegister(
		Media device,
		DirectoryNode dir,
		BusinessObject bo,
		long baseId)
		throws ObjectIOException {
		long id;
		setBusObjId(device, bo, baseId);
		id = bo.getId();
		setBusObjIdRecursive(device, bo, id);
		recursiveRegisterInternal(device, dir, bo);

	}

	private void recursiveRegisterInternal(
		Media device,
		DirectoryNode dir,
		BusinessObject bo)
		throws ObjectIOException {
		BusinessObject[] children = bo.getBusObjChildren();
		for (int j = 0; j < children.length; j++) {
			recursiveRegisterInternal(device, dir, children[j]);
		}
		register(device, dir, bo);
	}

	private void setBusObjIdRecursive(
		Media device,
		BusinessObject bo,
		long baseId)
		throws ObjectIOException {
		BusinessObject[] children = bo.getBusObjChildren();
		for (int j = 0; j < children.length; j++) {
			// System.out.println("!!!!! Setting children id's");
			setBusObjIdRecursive(device, children[j], baseId);
		}
		setBusObjId(device, bo, baseId);
	}

	/**
	 * Write an ObservationBlock and it's associated Target and OD to
	 * permanent media.  The in-memory copies remain available.
	 * 
	 * @param cb	CalibrationBlock or ObservationBlock to be written.
	 * 
	 * @exception ObjectIOException 			if there was an I/O error
	 * writing the object.
	 */
	public void putObHierarchy(CalibrationBlock cb) throws ObjectIOException {
		MultiObjSpec[] mos = toMultiObjSpecArray(cb);
		RegistryEntry oi = registry_.lookup(cb.getId());
		StorageManager stMgr = oi.getStorageManager();

		stMgr.multiStore(mos);

		for (int j = mos.length - 1; j >= 0; j--) {
			StorableObject sobj = mos[j].obj.getObject();
			RegistryEntry re = registry_.lookup(sobj.getId());

			re.isNew_ = false;
		}
	}
    
    
    
	protected MultiObjSpec[] toMultiObjSpecArray(CalibrationBlock cb)
		throws ObjectIOException {
		Vector specs = new Vector();

		Target tg =
			(cb instanceof ObservationBlock)
				? ((ObservationBlock) cb).getTarget()
				: null;
		ObservationDescription od = cb.getOd();
		ConstraintSet cs = null;
		if (cb instanceof ObservationBlock)
			cs = ((ObservationBlock) cb).getConstraintSet();

		if (tg != null) {
			specs.addElement(createMultiObjSpec(tg));
		}

		if (cs != null) {
			specs.addElement(createMultiObjSpec(cs));
		}

		specs.addElement(createMultiObjSpec(cb));

		if (od != null) {
			specs.addElement(createMultiObjSpec(od));
		}

		RegistryEntry oi = registry_.lookup(cb.getId());
		StorageManager stMgr = oi.getStorageManager();

		StorageManager.MultiObjSpec[] mos =
			new StorageManager.MultiObjSpec[specs.size()];
		specs.copyInto(mos);

		return mos;
	}

	private MultiObjSpec createMultiObjSpec(BusinessObject bo) {
		RegistryEntry oi = registry_.lookup(bo.getId());
		MultiObjSpec mos = new MultiObjSpec();

		mos.setNew(oi.isNew());
		mos.dir = oi.getDirectory();
		mos.setCls(bo.getClass());
		mos.obj = disassembleReferences(bo);

		return mos;
	}

	/**
	 * Get the root DirectoryNode object for a given Media object.
	 * 
	 * @param device The storage medium Local Store, RDBMS
	 * 
	 * @return Root Directory
	 * 
	 * @exception ObjectIOException
	 * 			thrown if the StorageManager for the media
	 * 			could not return the root directory
	 */
	public DirectoryNode getRoot(Media device) throws ObjectIOException {
		return getStorageManager(device).getRoot();
	}

	/**
	 * Set the id property of a (new) BusinessObject.
	 * 
	 * @param device	the medium on which the object is stored.
	 * @param bo	the BusinessObject with the id to be initialised.
	 * @param baseId	Id on which to base the new object id,
	 * 			(StorageManager-dependent)
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	protected void setBusObjId(Media device, StorableObject bo, long baseId)
		throws ObjectIOException {
		if (bo.getId() == 0) {
			bo.setId(getNewId(device, bo.getClass(), baseId));
		}
	}

	/**
	 * Asks the StorageManager to calculate a unique id.
	 *
	 * @param device	Media containing object for which id is needed.
	 * @param objType	type of the object to be given an id
	 * @return Unique id
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public long getNewId(Media device, Class objType)
		throws ObjectIOException {
		return getNewId(device, objType, 0);
	}

	/**
	 * Asks the StorageManager to calculate a unique id.
	 *
	 * @param device	Media containing object for which id is needed.
	 * @param objType	type of the object to be given an id
	 * @param baseId	Id on which to base the new object id,
	 * 			(StorageManager-dependent)
	 * @return Unique id
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	protected long getNewId(Media device, Class objType, long baseId)
		throws ObjectIOException {
		return getStorageManager(device).getNewId(objType, baseId);
	}

	/**
	 * Tests whether the specified object is in the Registry.
	 *
	 * @param id	id of a StorableObject
	 * 
	 * @return 		true if the specified object is in the Registry,
	 * 			false otherwise.
	 */
	public boolean isInRegistry(long id) {
		return (registry_.lookup(id) != null);
	}

	/**
	 * Add a link to this StorableObject into the Registry
	 * so that it can be found quickly when needed.  This is used to
	 * ensure that we only have one copy of an object in memory at
	 * any one time!
	 * 
	 * @param device	medium on which the object is stored.
	 * @param dir	directory in which the object is/should be stored.
	 * @param bo	StorableObject with the id to be initialised.
	 */
	protected void linkBusObj(
		Media device,
		DirectoryNode dir,
		boolean isNew,
		StorableObject bo) {
		StorageManager stMgr = getStorageManager(device);
		RegistryEntry oi = new RegistryEntry(bo, device, stMgr, dir, isNew);

		// oi.obj    = bo;
		// oi.stMgr  = stMgr;
		// oi.device = device;
		// oi.dir    = dir;
		// oi.isNew_  = isNew;

		registry_.add(bo.getId(), oi);
	}

	public RegistryEntry lookup(long id) {
		return registry_.lookup(id);
	}

	/**
	 * Move a directory within the specified storage medium.
	 *
	 * @param device	device on which the directory is stored
	 * @param node	directory to be moved
	 * @param toNode	the new parent directory
	 * 
	 * @return		True on success, false on failure
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 * 
	 * @exception ObjectNotFoundException
	 *			if the directory <code>node</code>was not found.
	 */
	public boolean moveDirectory(
		Media device,
		DirectoryNode node,
		DirectoryNode toNode)
		throws ObjectIOException, ObjectNotFoundException {
		StorageManager stMgr = getStorageManager(device);
		DirectoryNode parent = node.getParent();

		boolean movedOk = stMgr.moveDirectory(node, toNode);

		if (movedOk) {
			fireDirectoryEvent(
				DirectoryEvent.DIRECTORY_MOVED,
				device,
				node,
				parent);
		}

		return movedOk;
	}

	protected void dequeueObject(StorableObject sobj) {
	}

	public void queueObject(StorableObject sobj) {
	}

	/**
	 * Move an object within the specified storage medium.
	 *
	 * @param device	device on which the directory is stored
	 * @param id	id of the object to be moved
	 * @param objType	type of the object to be moved
	 * @param toNode	the new parent directory
	 * 
	 * @return		True on success, false on failure
	 * 
	 * @exception ObjectIOException
	 * 			if an I/O error occurred during the operation
	 * 
	 * @exception ObjectNotFoundException
	 *			if the directory <code>fromNode</code>was not found.
	 */
	public void moveBusObj(
		Media device,
		long id,
		Class objType,
		DirectoryNode toNode)
		throws ObjectNotFoundException, ObjectIOException {
		StorageManager stMgr = getStorageManager(device);
		RegistryEntry oi = registry_.lookup(id);
		DirectoryNode fromNode = null;
		StorableObject sobj = null;

		if (oi != null) {
			fromNode = oi.getDirectory();
			sobj = oi.getObject();
			if (fromNode == toNode) {
				return; // Nothing to do, just return
			}
			dequeueObject(oi.getObject());
		}

		try {
			// NB: This is effectively an "if" statement
			//     If find() doesn't throw an exception then move()
			fromNode = stMgr.find(id, objType);
			stMgr.move(id, objType, fromNode, toNode);
		} catch (ObjectNotFoundException e) {
			// ignore -- if not found, the object is new and not yet saved.
			//           We explicitly save it below, so all will be well.
		}

		if (fromNode == null) {
			throw new ObjectNotFoundException("Unknown object: " + id);
		}

		if (oi != null) {
			oi.setDirectory(toNode);
			queueObject(oi.getObject());
		}

		fireObjectEvent(
			ObjectEvent.OBJECT_MOVED,
			device,
			fromNode,
			toNode,
			sobj,
			id,
			objType);
	}
    
    
 

	/**
	 * Delet a Business Object. Remove the object from the Registry.
	 * 
	 * @param device	the medium on which the object is deleted
	 * @param dir	the directory in which the object is/should be deleted.
	 * @param bo	the StorableObject with the id to be initialised.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void deleteBusObj(Media device, long id, Class objType)
		throws ObjectNotFoundException, ObjectIOException {
		StorageManager stMgr = getStorageManager(device);
		RegistryEntry oi = registry_.lookup(id);
		StorableObject sobj = null;
		DirectoryNode node = null;

		if (oi != null) {
			node = oi.getDirectory();
			sobj = oi.getObject();
		} else {
			node = stMgr.find(id, objType);
		}

		stMgr.delete(node, id, objType);
		registry_.remove(id);

		fireObjectEvent(
			ObjectEvent.OBJECT_REMOVED,
			device,
			null,
			node,
			sobj,
			id,
			objType);
	}

	/**
	 * Add the Business Object to the Registry.         
	 * 
	 * @param device	the medium on which the object is to be stored
	 * @param sobj	the BusinessObject to be added.
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void register(Media device, StorableObject sobj)
		throws ObjectIOException {
		register(device, getRoot(device), sobj, 0);
	}

	/**
	 * Add the Business Object to the Registry.         
	 * 
	 * @param device	the medium on which the object is to be stored
	 * @param sobj	the BusinessObject to be added.
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void register(Media device, StorableObject sobj, long baseId)
		throws ObjectIOException {
		register(device, getRoot(device), sobj, baseId);
	}

	/**
	 * Add a Business Object. Add the Business Object to the Registry.         
	 * 
	 * @param device	the medium on which the object is to be stored
	 * @param dir	the directory in which the object is/should be stored.
	 * @param sobj	the BusinessObject to be added.
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void register(Media device, DirectoryNode dir, StorableObject sobj)
		throws ObjectIOException {
		register(device, dir, sobj, 0);
	}

	/**
	 * Add a New Storable Object.
	 * 
	 * @param device	the medium on which the object is to be stored
	 * @param dir	the directory in which the object is/should be stored.
	 * @param sobj	the StorableObject to be added.
	 * @param baseId	Id on which to base the new object id,
	 * 			(StorageManager-dependent)
	 * 
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void register(
		Media device,
		DirectoryNode dir,
		StorableObject sobj,
		long baseId)
		throws ObjectIOException {
		setBusObjId(device, sobj, baseId);
		if (sobj instanceof BusinessObject) {
			((BusinessObject) sobj).setOwnerId(Config.getCfg().getUserName()); 
		}
		linkBusObj(device, dir, true, sobj);
		fireObjectEvent(
			ObjectEvent.OBJECT_CREATED,
			device,
			null,
			dir,
			sobj,
			sobj.getId(),
			sobj.getClass());
	}

	/**
	 * Write a StorableObject to its permanent media.  The in-memory
	 * copy remains available.
	 * 
	 * @param sobj	Object to be stored.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public void putBusObj(StorableObject sobj)
		throws ObjectNotFoundException, ObjectIOException {
		RegistryEntry oi = registry_.lookup(sobj.getId());
		UnpackedStorableObject objSpec;
        
		if (oi == null) { 
			throw new ObjectNotFoundException(
				"object not in memory: " + sobj.getId());
		}       
		objSpec = disassembleReferences(oi.getObject());
		StorageManager stMgr = oi.getStorageManager();
stdlog_.debug("////////////////////// Using the following id: " + sobj.getId());
		if (oi.isNew()) {
			stMgr.write(
				oi.getDirectory(),
				sobj.getId(),
				sobj.getClass(),
				objSpec);
		} else {
			stMgr.update(
				oi.getDirectory(),
				sobj.getId(),
				sobj.getClass(),
				objSpec);
		}
        oi.setLocked(false);
		oi.setNew(false);
	}

	/**
	 * Forget a BusinessObject, ie remove it from the Registry.
	 * When using this method, be careful that there no classes
	 * that are holding a reference to it.  If you try (eg) to write
	 * the BusinessObject after removing it from the Registry, an
	 * exception wil be thrown.
	 * 
	 * <p>
	 * If the object is not currently in the Registry, the method
	 * just returns.
	 * 
	 * @param id	Id of object to be forgotten
	 * 
	 */
	public void forgetBusObj(Media device, long id, Class objType) {
		RegistryEntry oi = registry_.lookup(id);

		if (oi != null) {
			DirectoryNode node = oi.getDirectory();
			StorableObject sobj = oi.getObject();
			dequeueObject(oi.getObject());
			registry_.remove(id);
			fireObjectEvent(
				ObjectEvent.OBJECT_FORGOTTEN,
				device,
				null,
				node,
				sobj,
				id,
				objType);
		}
	}

	/**
	 * Refresh a BusinessObject instance from external media.
	 * This is useful in case the version in a database has changed
	 * and we want the updated version, rather that the local cached copy.
	 * 
	 * @param device	External medium on which the object is stored
	 * @param id	Id of object to be refreshed
	 * @param objType	Type of object to be refreshed
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 *			if there was an I/O error refreshing the object.
	 */
	public void refreshBusObj(Media device, long id, Class objType)
		throws ObjectNotFoundException, ObjectIOException {
		if (!BusinessObject.class.isAssignableFrom(objType)) {
			throw new ObjectIOException(
				"refreshBusObj: cannot refresh this type of object: "
					+ objType.getName());
		}

		if (!isInRegistry(id)) {
			return; // Not in Registry -- nothing to refresh
		}

		StorageManager stMgr = getStorageManager(device);
		DirectoryNode node = stMgr.find(id, objType);
		UnpackedStorableObject uso = stMgr.read(node, id, objType);
		BusinessObject dbBO = (BusinessObject) uso.getObject();
		BusinessObject localBO =
			(BusinessObject) getBusObj(device, id, objType);

		try {
			String[] props = dbBO.simpleProperties();
			dbBO.copyProperties(props, localBO);
		} catch (NoSuchMethodException e) {
			throw new ObjectIOException(
				"refreshBusObj: NoSuchMethod: " + e.getMessage());
		} catch (InvocationTargetException e) {
			throw new ObjectIOException(
				"refreshBusObj: InvocationTargetException:"
					+ e.getTargetException().getMessage());
		} catch (IllegalAccessException e) {
			throw new ObjectIOException(
				"refreshBusObj: IllegalAccess: " + e.getMessage());
		}

	}

	/**
	    * Refresh a BusinessObject and all its children business objec from 
	    * external media.
	    * This is useful in case the version in a database has changed
	    * and we want the updated version, rather that the local cached copy.
	    * 
	    * @param device	External medium on which the object is stored
	    * @param id	Id of object to be refreshed
	    * @param objType	Type of object to be refreshed
	    * 
	    * @exception ObjectNotFoundException
	    *			if the object was not found.
	    * @exception ObjectIOException
	    *			if there was an I/O error refreshing the object.
	    */
	public void refreshBusObjRecursive(Media device, long id, Class objType)
		throws ObjectNotFoundException, ObjectIOException {

		if (!isInRegistry(id)) {
			return; // Not in Registry -- nothing to refresh
		}

		BusinessObject localBO =
			(BusinessObject) getBusObj(device, id, objType);

		Stack stk = new Stack();
		stk.push(localBO);
		while (!stk.isEmpty()) {

			BusinessObject busObj = (BusinessObject) stk.pop();
			// 	    System.out.println(Config.getCfg().getClassFromId(busObj.getId()));

			BusinessObject[] kids = busObj.getBusObjChildren();
			for (int j = 0; j < kids.length; j++) {
				stk.push(kids[j]);
			}
			refreshBusObj(
				device,
				busObj.getId(),
				Config.getCfg().getClassFromId(busObj.getId()));
		}
	}

	/**
	 * Read a StorableObject from its permanent media, link it into the
	 * StorableObject infrastructure, etc and return it.
	 * 
	 * @param device	the medium on which the StorableObject is stored.
	 * @param id	the id of the StorableObject to store.
	 * @param objType	the type of the object to store.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 * 
	 * @return StorableObject
	 */
	public StorableObject getBusObj(Media device, long id, Class objType)
		throws ObjectNotFoundException, ObjectIOException {
		RegistryEntry oi = registry_.lookup(id);

		if (oi != null) {
            stdlog_.debug("!!!!!!!!!! Returning from Registry " + id);
			return oi.getObject();
		}

		StorageManager stMgr = getStorageManager(device);
		DirectoryNode node = stMgr.find(id, objType);
		UnpackedStorableObject objSpec = stMgr.read(node, id, objType);

		if (objSpec == null) {
			org.eso.ohs.core.utilities.DebugUtils.stackTrace("SHOULDN'T HAPPEN");
			return null;
		}

		StorableObject sobj = (StorableObject) objSpec.getObject();

		reassembleReferences(device, objSpec);

		linkBusObj(device, node, false, sobj);

		fireObjectEvent(
			ObjectEvent.OBJECT_READ,
			device,
			null,
			node,
			sobj,
			sobj.getId(),
			sobj.getClass());
		return sobj;
	}

	/**
	 * Restore the saved the links in a StorableObject that has been read
	 * from external media.
	 * 
	 * <p>
	 * The objSpec parameter must follow the format as results from
	 * StorageManager.read().
	 * 
	 * <p>
	 * <b>Note:</b> XXX after migration to JDK 1.2, this method should
	 *              ignore the setX/getX methods and directly set
	 *              the transient fields.
	 * 
	 * @param device	the medium on which the StorableObject is stored.
	 * @param objSpec	the array of objects read from the storage medium.
	 * @param objType	the type of the object to store.
	 * 
	 * @exception ObjectNotFoundException
	 *			if the object was not found.
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 * 
	 * @see StorageManager#read
	 */
	protected void reassembleReferences(
		Media device,
		UnpackedStorableObject objSpec)
		throws ObjectIOException, ObjectNotFoundException {
		StorableObject sobj = objSpec.getObject();
		Enumeration enum = objSpec.getProperties();
		long lastMod =
			sobj instanceof BusinessObject
				? ((BusinessObject) sobj).getLastModifiedDate()
				: -1;

		// ASSERT objSpec.length is odd

		while (enum.hasMoreElements()) {
			String propName = (String) enum.nextElement();
			long nestedId = objSpec.getNestedObjId(propName);
			Object nestedObj = null;

			try {
				Class cl = sobj.getClass();
				Method getMeth = cl.getMethod("get" + propName, new Class[] {
				});
				Method setMeth =
					cl.getMethod(
						"set" + propName,
						new Class[] { getMeth.getReturnType()});

				if (nestedId != 0L) {
					// XXX It should which dbase contains an objType
					Class objType = getMeth.getReturnType();
					Media dev = device;

					if (objType == ObservingRun.class
						&& device.equals(Media.PHASE2_DB)) {
						dev = Media.PHASE1_DB;
					}

					nestedObj = getBusObj(dev, nestedId, objType);
				}

				setMeth.invoke(sobj, new Object[] { nestedObj });
			} catch (IllegalAccessException e) {
				stdlog_.error(e);
			} catch (InvocationTargetException e) {
				stdlog_.error(e.getTargetException());
			} catch (NoSuchMethodException e) {
				stdlog_.error(e);
				throw new IllegalArgumentException("corrupt data: " + propName);
			}
		}

		if (lastMod >= 0) {
			((BusinessObject) sobj).setLastModifiedDate(lastMod);
		}
	}

	/**
	 * Save the links in a StorableObject that is to be written
	 * to external media.
	 * 
	 * <p>
	 * <b>Note:</b> XXX after migration to JDK 1.2, this method should
	 *              ignore the setX/getX methods and directly set
	 *              the transient fields.
	 *
	 * @param bo Use the StorableObject to create an array of Objects 
	 *           that contains the field names and method names of the
	 *           Objects that link to these objects.
	 * @return Object[] A generic list of objects that contain the methods
	 *                  and fields from the linked objects.
	 *
	 */
	protected UnpackedStorableObject disassembleReferences(StorableObject sobj) {
		UnpackedStorableObject uso = new UnpackedStorableObject(sobj);

		try {
			Class cl = sobj.getClass();
			Method[] meths = cl.getMethods();
			Class boClass = BusinessObject.class;

			for (int j = 0; j < meths.length; j++) {
				Method m = meths[j];
				Class retType = m.getReturnType();

				// Look for a getX() method returning a subclass
				// of StorableObject.
				if (m.getName().startsWith("get")
					&& boClass.isAssignableFrom(retType)
					&& m.getParameterTypes().length == 0) {

					BusinessObject field =
						(BusinessObject) m.invoke(sobj, NO_ARGS);

					uso.addNestedObjId(
						m.getName().substring(3),
						field == null ? 0 : field.getId());
				}
			}
		} catch (IllegalAccessException e) {
			stdlog_.error(e);
		} catch (InvocationTargetException e) {
			stdlog_.error(e);
		}

		return uso;
	}

	/**
	 * Creates a directory on external media.
	 *
	 * @param device	Reference to the selected media. 
	 * @param parent	Parent of directory to be created on the selected media 
	 * @param id	id of the new directory
	 * 
	 * @return True if directory created successfully otherwise false      
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public long createDirectory(Media device, DirectoryNode parent, long id)
		throws ObjectIOException {
		StorageManager stMgr = getStorageManager(device);

		if (stMgr.createDirectory(parent, id)) {
			fireDirectoryEvent(
				DirectoryEvent.DIRECTORY_CREATED,
				device,
				parent.getDirectory(id),
				parent);
			return id;
		} else {
			return NO_OBJECT;
		}
	}
    
    


	/**
	 * Test for whether the user is allowed to delete the directory.
	 * Used by deleteDirectory().  Hook for subclasses.
	 * 
	 * @param device A reference to the selected media. 
	 * @param node The directory to be deleted on the selected media
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	protected boolean canDeleteDirectory(Media device, DirectoryNode node)
		throws ObjectIOException {
		return node.getChildCount() == 0
			&& !registry_.hasObjectsInDirectory(node);
	}

	/**
	 * Hook to allow an application-specific subclass to
	 * prepare a directory for deletion, eg cleaning out
	 * application-specific data contained in the directory.
	 * 
	 * @param device A reference to the selected media. 
	 * @param node The directory to be deleted on the selected media
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	protected void prepareDirectoryForDeletion(
		Media device,
		DirectoryNode node)
		throws ObjectIOException {
		// hook for subclasses
	}

	/**
	 * Deletes a directory on external media.
	 *
	 * @param device A reference to the selected media. 
	 * @param node The directory to be deleted on the selected media
	 * @return True if directory deleted successfully otherwise false    
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */
	public boolean deleteDirectory(Media device, DirectoryNode node)
		throws ObjectIOException {

		DirectoryNode parent = node.getParent();

		if (!canDeleteDirectory(device, node)) {
            stdlog_.debug("Cannot delete direcctory");
			return false;
		}

		prepareDirectoryForDeletion(device, node);

		if (getStorageManager(device).deleteDirectory(node)) {

			fireDirectoryEvent(
				DirectoryEvent.DIRECTORY_REMOVED,
				device,
				node,
				parent);
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Gets the lists of different object types(targets,ob's,od's) for a given directory
	 * 
	 *
	 * @param device A reference to the selected media. 
	 * @param node The directory that is used to select the appropiate
	 *             directory
	 * @return Returns a list of ObjectFileInfo for a given directory
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 */

	public Summary[] listObjects(Media device, DirectoryNode node)
		throws ObjectIOException {
		return getStorageManager(device).listObjects(node);
	}

	/**
	 * Returns the ObservingRun object with the given id.
	 *
	 * @param device	Reference to the selected media. 
	 * @param node	Directory containing the ObservingRun
	 * @return Returns a list of ObjectFileInfo for a given directory
	 *
	 * @exception ObjectIOException
	 *			if there was an I/O error writing the object.
	 * @exception ObjectNotFoundException
	 *			if the ObservingRun object was not found.
	 */
	public ObservingRun getObservingRun(long id)
		throws ObjectIOException, ObjectNotFoundException {
		StorageManager stMgr = getStorageManager(Media.PHASE1_DB);
		UnpackedStorableObject objs = stMgr.read(null, id, ObservingRun.class);
		ObservingRun obsRun = (ObservingRun) objs.getObject();

		return obsRun;
	}

	/**
	 * Adds a new listener to be notified of created/deleted directories
	 * 
	 * @param lnr the listener
	 */
	public void addDirectoryListener(DirectoryListener lnr) {
		dirListeners.add(DirectoryListener.class, lnr);
	}

	/**
	 * Removes a listener, which will no longer be notified of changes.
	 * 
	 * @param lnr the listener
	 */
	public void removeDirectoryListener(DirectoryListener lnr) {
		dirListeners.remove(DirectoryListener.class, lnr);
	}

	/**
	 * Inform listeners of a change to the directory hierarchy.
	 * @param evtType type of Event generated 
	 * @param device the currently selected storage device
	 * @param dir The directory which has been created or removed
	 */
	protected void fireDirectoryEvent(
		int evtType,
		Media device,
		DirectoryNode dir,
		DirectoryNode parent) {
		DirectoryEvent objEvent =
			new DirectoryEvent(evtType, device, dir, parent);
		Object[] listeners = dirListeners.getListenerList();

		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			Object lnr = listeners[i + 1];
			if (listeners[i] == DirectoryListener.class) {
				if (evtType == DirectoryEvent.DIRECTORY_CREATED) {
					((DirectoryListener) lnr).directoryAdded(objEvent);
				} else if (evtType == DirectoryEvent.DIRECTORY_MOVED) {
					((DirectoryListener) lnr).directoryMoved(objEvent);
				} else if (evtType == DirectoryEvent.DIRECTORY_CHANGED) {
					((DirectoryListener) lnr).directoryChanged(objEvent);
				} else {
					((DirectoryListener) lnr).directoryRemoved(objEvent);
				}
			}
		}
	}
    
    /**
     * Get the database id for an object.
     * @param obj
     * @return The database id.
     */
    
    public long getDbaseId(BusinessObject obj) {
        Media device =
            ObjectManager.getObjectManager().lookup(obj.getId()).getMedia();
    
        long id;
        if (device.equals(Media.DBASE))
            id = obj.getDbaseId();
        else {
            long dbIdTemp = obj.getDbaseId();
            id =
                (dbIdTemp != 0)
                    ? Config.getCfg().uniqueToTableId(dbIdTemp)
                    : obj.getId();
        }
        return id;
    }


	/**
	 * Adds a new listener to be notified of created/read/deleted objects
	 * 
	 * @param lnr The listener
	 */
	public void addObjectListener(ObjectListener lnr) {
		objectListeners.add(ObjectListener.class, lnr);
	}

	/**
	 * Removes a listener, which will no longer be notified of changes.
	 * 
	 * @param lnr The listener
	 */
	public void removeObjectListener(ObjectListener lnr) {
		objectListeners.remove(ObjectListener.class, lnr);
	}

	/**
	 * Inform listeners of a change to the Registry of StorableObjects
	 *
	 * @param evtType types of ObjecyEvent
	 * @param dir directory that the event will take place on.
	 * @param sobj reference to storable object
	 */
	protected void fireObjectEvent(
		int evtType,
		Media device,
		DirectoryNode fromDir,
		DirectoryNode toDir,
		StorableObject sobj,
		long id,
		Class objType) {
		ObjectEvent objEvent =
			new ObjectEvent(evtType, device, fromDir, toDir, sobj, id, objType);
		Object[] listeners = objectListeners.getListenerList();

		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			Object lnr = listeners[i + 1];
			if (listeners[i] == ObjectListener.class) {
				if (evtType == ObjectEvent.OBJECT_READ
					|| evtType == ObjectEvent.OBJECT_CREATED) {
					((ObjectListener) lnr).objectAdded(objEvent);
				} else if (evtType == ObjectEvent.OBJECT_MOVED) {
					((ObjectListener) lnr).objectMoved(objEvent);
				} else if (evtType == ObjectEvent.OBJECT_REMOVED) {
					((ObjectListener) lnr).objectRemoved(objEvent);
				} else if (evtType == ObjectEvent.OBJECT_FORGOTTEN) {
					((ObjectListener) lnr).objectRemovedFromRegistry(objEvent);
				}
			}
		}
	}

	/**
	 * A container class for several fields and used only with the
	 * inMemory table (a.k.a. the Registry).
	 * 
	 * <b>NOTE:</b>  Some of these methods should perhaps be synchronized
	 *               in future to protect against concurrent access but
	 *		     at the moment there seems no need.
	 */
	public static class RegistryEntry {
		/** The object itself */
		private StorableObject sobj_;

		/** External media upon which the object is stored */
		private Media device_;

		/** StorageManager used to update the object on external media */
		private StorageManager stMgr_;

		/** Logical directory in which the object is stored on the media */
		private DirectoryNode dir_;

		/** No copy of the object exists on the medium? Important for DB */
		private boolean isNew_;
        
        
        private boolean locked=false;
        

		public RegistryEntry(
			StorableObject sobj,
			Media device,
			StorageManager stMgr,
			DirectoryNode dir,
			boolean isNew) {

			if (dir == null) {
				throw new NullPointerException("Bad Registry entry: no dir");
			}

			if (sobj == null) {
				throw new NullPointerException("Bad Registry entry: no object");
			}

			sobj_ = sobj;
			device_ = device;
			stMgr_ = stMgr;
			dir_ = dir;
			isNew_ = isNew;
		}

		/**
		 * Return a refencet to the Storable object that the 
		 * RegistryEntry refers to
		 * 
		 * @return StorableObject
		*/

		public StorableObject getObject() {
			return sobj_;
		}

		/**
		 * The Media where the Registry Entry resides
		 * @return Media Database,Local Storage
		 */

		public Media getMedia() {
			return device_;
		}
        
        
        

		/**
		 * The storage manager that is used to store the 
		 * BusinessObject to the database         
		 */

		public StorageManager getStorageManager() {
			return stMgr_;
		}

		/**
		 * Set the directory associated with the Storable Object
		 * @return dir Directory Node
		 */

		public void setDirectory(DirectoryNode dir) {
			dir_ = dir;
		}

		/**
		 * Return the directort associated with the Storable Object
		 * This does not apply to Dbase Objects
		 * @return DirectortNode for business Objects
		 */

		public DirectoryNode getDirectory() {
			return dir_;
		}

		/**
		 * Set the new value for the Registry Entry 
		 * @param isNew 
		 */

		public void setNew(boolean isNew) {
			isNew_ = isNew;
		}

		/**
		 * If the Registry Entry is already in the registry         
		 * @return isNew
		 */

		public boolean isNew() {
			return isNew_;
		}

        /**
         * Returns the locked.
         * @return boolean
         */
        public boolean isLocked() {
            return locked;
        }

        /**
         * Sets the locked.
         * @param locked The locked to set
         */
        public void setLocked(boolean locked) {
            this.locked = locked; 
        }

	}

	/**
	 * The Registry is a table of cached StorableObject instances.
	 * There will only ever be one Registry at any time but it is
	 * not a singleton because it is contained within the ObjectManager,
	 * which is itself a singleton.
	 */
	public static class Registry {
        
        
          
            /* public static HashMap alwaysInMemory = new HashMap();
        
        static {
            alwaysInMemory.put(ObservationBlock.class,Boolean.FALSE);         
            alwaysInMemory.put(ConstraintSet.class,Boolean.FALSE);
            alwaysInMemory.put(ObservationBlock.class,Boolean.FALSE);
            alwaysInMemory.put(CalibrationBlock.class,Boolean.FALSE);
            alwaysInMemory.put(Target.class,Boolean.FALSE);
            alwaysInMemory.put(Folder.class,Boolean.TRUE);
            alwaysInMemory.put(ObservationDescription.class,Boolean.FALSE);
            alwaysInMemory.put(ObservingRun.class,Boolean.TRUE);
            alwaysInMemory.put(Mask.class,Boolean.FALSE);
            }*/
        
        
		/**
		 * References to ObjectInfo records for every StorableObject that
		 * is currently in memory.  The table is keyed by the object's id
		 * wrapped in a java.lang.Long.
		 * 
		 * <p>
		 * Using this table, we can ensure that at most one instance of
		 * any StorableObject is kept in memory at any one time.
		 */
        
		 //protected final Map inMemory_ =  Collections.synchronizedMap(new ReferenceMap(ReferenceMap.HARD,ReferenceMap.HARD,1000,0.5f));
                protected final Hashtable inMemory_ =  new Hashtable();


		/**
		 * Return a RegistryEntry object for the given id, or
		 * null if not found.
		 * 
		 * @param id	id of the StorableObject to be searched for
		 * 
		 * @return	RegistryEntry for the object, or null if not found
		 */
		public RegistryEntry lookup(long id) {
			return lookup(new Long(id));
		}

		/**
		 * Return a RegistryEntry object for the given id (wrapped in a Long),
		 * or null if not found.
		 * 
		 * @param idObj	id of the StorableObject to be searched for,
		 * 		wrapped in a java.lang.Long
		 * 
		 * @return	RegistryEntry for the object, or null if not found
		 */
		public synchronized RegistryEntry lookup(Long idObj) {
            return (RegistryEntry)inMemory_.get(idObj);
		}

		/**
		 * Return a RegistryEntry object for the given id (wrapped in a Long),
		 * or null if not found.
		 * 
		 * @param idObj	id of the StorableObject to be searched for,
		 * 		wrapped in a java.lang.Long
		 * 
		 * @return	RegistryEntry for the object, or null if not found
		 */
		public synchronized void add(long id, RegistryEntry oi) {
			if (oi == null) {
				throw new NullPointerException("Bad Registry entry: " + oi);
			}
			if (lookup(id) == null) {
				inMemory_.put(new Long(id),oi);
			}
		}

		/**
		 * Remove the table entry for the given id, if it exists.
		 * 
		 * @param id	id of the table entry to be removed
		 * 
		 * @return	RegistryEntry for the object, or null if not found
		 */
		public synchronized RegistryEntry remove(long id) {
			return (RegistryEntry) inMemory_.remove(new Long(id));
		}

		/**
		 * Returns true if there are objects in the Registry that are
		 * recorded as being in this directory.
		 * 
		 * @param node	DirectoryNode for the directory to be checked
		 * 
		 * @return	true if there are one or more entries for this
		 * 		directory, false otherwise
		 */
		public synchronized boolean hasObjectsInDirectory(DirectoryNode node) {
            return inMemory_.containsValue(node);
		}
	}
    
    /* private static class LRULockMap extends LRUMap {
        
        
        public LRULockMap(int size) {
            super(size);
        }
                
        protected void removeLRU() {
            Object key = getFirstKey();
            // be sure to ca Object key = getFirstKey();
            // be sure to call super.get(key), or you're likely to 
            // get infinite promotion recursion
            stdlog_.debug("Maximum Size " + getMaximumSize());
            Object value = super.get(key);
            RegistryEntry oi = (RegistryEntry)value;
            long id = oi.getObject().getId();
            SimpleStorableObject obj = (SimpleStorableObject) oi.getObject();
            stdlog_.debug("Id = " + id);
            stdlog_.debug("Class = " + obj.getClass());
            Class c = obj.getClass();
            Boolean inMemory = (Boolean)Registry.alwaysInMemory.get(c);
            remove(key);            
            processRemovedLRU(key,value);
        }
       
        } */   
}

