package org.eso.ohs.persistence;

import javax.swing.event.EventListenerList;

import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.StorableObject;

/**
 * This class manages the in-memory copies of all StorableObjects.
 * When a reference to some StorableObject is requested by a client,
 * it first looks it up in a table to see if the object is already
 * in memory and, if so, returns it.  This ensures that only one
 * copy of any StorableObject is in memory at any one time, which will
 * eventually be written to the external media.  Thus an object cannot
 * appear to have several values at once.
 * 
 * <p>
 * If the request object is not in memory, it is retrieved from the
 * associated external media, added to the inMemory table and returned.
 * As part of the retrieval process, any StorableObjects referred to by
 * this object are also read into memory and referred to correctly.
 * 
 * <p>
 * For correct operation, there should be at most one instance of this
 * object.  Thus it is a singleton and can only be instantiated using
 * the static PhrsObjectManager.getObjectManager() method.  Each time this
 * method is called it will return the <em>same</em> instance of the
 * PhrsObjectManager.
 * 
 * <p>
 * $Id: PhrsObjectManager.java,v 1.11 2004/02/17 14:31:56 tcanavan Exp $
 * <br>* @author Bob Kemp
 */
public class PhrsObjectManager extends AbstractObjectManager{

  /** Invalid object id.  Used to indicate that there is no such object */
 
    /* Support for handling object events */
    protected EventListenerList objectListeners;

    /* Support for handling directory events */
    protected EventListenerList dirListeners;


    /**
     * Construct an instance of the PhrsObjectManager.
     * The constructor is protected, so that we can ensure there is
     * only one instance of the PhrsObjectManager, i.e. it's a singleton.
     */
    protected PhrsObjectManager() {
	super();
	dirListeners  = new EventListenerList();     
    }



    /**
     * Return the one and only instance of the PhrsObjectManager.
     * 
     * @return Singleton PhrsObjectManager instance.
     */
    public static PhrsObjectManager getObjectManager() {
	if (objMgr_ == null) {
	    objMgr_ = new PhrsObjectManager();
	}
	return (PhrsObjectManager)objMgr_;
    }




    public void forgetBusObj(BusinessObject obj,Class objType) {
	long id = obj.getId();
	super.forgetBusObj(Media.PHASE1_DB,id,objType);
    }


    public void forgetBusObj(long id, Class objType) {
	long objMgrId = Config.getCfg().tableToUniqueId(id, objType);
	super.forgetBusObj(Media.PHASE1_DB,objMgrId,objType);
    }


    /**
     * Startup.
     * 
     * @see #restart
     * @see #shutdown
     */
    public void startup() {
	setObjectManager(new PhrsObjectManager());
    }



    /**
     * Shutdown hook for subclasses.
     * 
     * @see #startup
     * @see #restart
     */
    public void shutdown() {
	objMgr_ = null;
    }


    public StorableObject[]  register(StorableObject[] sobjArray)throws  ObjectIOException{
	return super.register(Media.PHASE1_DB,sobjArray);
    }

    public StorableObject[]  registerPersistence(StorableObject[] sobjArray)throws  ObjectIOException{
	return super.registerPersistence(Media.PHASE1_DB,sobjArray);
    }


    public StorableObject register(StorableObject sobj) throws  ObjectIOException {
	return super.register(Media.PHASE1_DB,sobj);
    }

    public StorableObject registerPersistence(StorableObject sobj) throws  ObjectIOException {
	return super.registerPersistence(Media.PHASE1_DB,sobj);
    }


  public StorableObject getBusObjNoRecursive(long id, Class objType)
				throws ObjectNotFoundException,
					ObjectIOException {
      return super.getBusObjNoRecursive(Media.PHASE1_DB,
					Config.getCfg().tableToUniqueId(id, objType),
					objType);
 }


    public void deleteBusObj(long id,
			     Class objType)
	throws ObjectNotFoundException,  ObjectIOException {
	super.deleteBusObj(Media.PHASE1_DB,
			   id,
			   objType);
 }


    public BusinessObject refreshBusObj(long id, Class objType)
	throws ObjectNotFoundException, ObjectIOException {
	BusinessObject bo = super.refreshBusObj(Media.PHASE1_DB,
			   Config.getCfg().tableToUniqueId(id, objType),
			   objType);
			   return bo;
    }
}


