package org.eso.ohs.phase2.apps.p2pp;

import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.util.Stack;
import java.util.Vector;

import javax.swing.JMenu;

import org.eso.ohs.core.gui.baseaction.IncrementalAction;
import org.eso.ohs.core.gui.baseaction.IncrementalActionAutomaton;
import org.eso.ohs.core.gui.widgets.ErrorMessages;
import org.eso.ohs.core.utilities.FileIOUtils;
import org.eso.ohs.core.utilities.ObjUtils;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.VisitorAdaptor;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.p2pp.Folder;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectNotFoundException;

/**
 * Action class to move a business object to the chosen directory.
 * 
 * <p>
 * $Id: MoveBOMenuHandler.java,v 1.17 2005/02/11 10:31:50 tcanavan Exp $
 */

public class MoveBOMenuHandler extends HierarchicalMenu {

    /** folder view in this window. */
    private FolderGridView folderView_;

    /**
     * Constructs a move BusinessObject menu handler.
     * 
     * @param device      medium containing the parent directory.
     * @param folderView  GUI view of (summary information in) the folder.
     * @param menuParent  JMenu containing the hierarchical submenu
     */
    public MoveBOMenuHandler(
        Media device,
        FolderGridView folderView,
        HierarchyView hView,
        JMenu menuParent) {
        super(device, menuParent, hView, "Move");

        folderView_ = folderView;
    }

    /**
     * Handler method for "Move A Business Object" menu item.
     * 
     * @param ev	event giving details of the requested action
     */
    public void actionPerformed(ActionEvent ev) {
        Persistence objMgr = Persistence.getInstance();
        DirectoryNode newParent = getNodeFromMenuItem(ev.getSource());
        Long[] idLongs = folderView_.getAllSelected();
        Class objType = folderView_.getDisplayedObjectClass();
        Vector newLongs = new Vector();
        for (int i = idLongs.length - 1; i >= 0; i--) {
            newLongs.addElement(idLongs[i]);
        }
        Long[] newIds = new Long[newLongs.size()];
        newLongs.copyInto(newIds);
        Folder folder = folderView_.getFolder();
        if (!checkSameInstrument(folder.getNode(), newParent)) {
            return;
        }
        MenuGroups frm = (MenuGroups) folderView_.getTopLevelAncestor();

        new IncrementalActionAutomaton(
            new IncrementalMover(
                frm,
                folder.getMedia(),
                newParent,
                objType,
                newIds),
            frm,
            "Moving objects...");

    }

    /**
     * Class to incrementally move a group of business objects
     */
    public class IncrementalMover implements IncrementalAction {

        /** cached reference to Persistence */
        private Persistence objMgr = Persistence.getInstance();

        /** array of ids of objects to be moved */
        private Long[] idLongs_;

        /** Type of business object to be moved */
        private Class objType_;

        /** Media containing the objects to be moved */
        private Media device_;

        /** Directory to which the BOs should be moved */
        private DirectoryNode newParent_;

        /** Window on which to base all dialogue boxes */
        private Frame baseFrame_;

        /** 
         * Constructor.
         * 
         * @param frm	    	base frame for any dialog boxes, eg error msgs
         * @param device    	media containing the objects to be moved
         * @param newParent	directory to which the BOs should be moved
         * @param objType	type of business object to be moved
         * @param idLongs	array of ids of objects to be moved
         */
        public IncrementalMover(
            Frame frm,
            Media device,
            DirectoryNode newParent,
            Class objType,
            Long[] idLongs) {
            baseFrame_ = frm;
            device_ = device;
            newParent_ = newParent;
            objType_ = objType;
            idLongs_ = idLongs;
        }

        public int getNumStepsRemaining(int counter) {
            return ObjUtils.howManyNotNull(idLongs_, counter);
        }

        public void startOfAction() {
        }

        public void endOfAction() {
        }

        /**
         * Method to move the n'th element of obs, where n = counter.
         * 
         * @param counter   counter to select the element of idLong.
         * @return	    true if more objects should be moved
         *		    and false otherwise.
         */
        public void performStep(int counter) {
            if (idLongs_[counter] != null) {
                doMove(idLongs_[counter].longValue());
            }
        }

        /**
         * Utility method to move the specified object.
         * 
         * @param id	    id of the object to move.
         */
        public void doMove(long id) {
            BusinessObject bo = null;

            try {
                Folder folder = FolderFactory.getFolder(device_, newParent_);
                String destFolderPath =
                    Persistence.getInstance().getRootFolderPath(
                        folder.findObsRun());
                bo = (BusinessObject) objMgr.getBusObj(device_, id, objType_);
                Stack stk = new Stack();
                stk.push(bo);
                CalibrationBlock cb = null;
                while (!stk.isEmpty()) {
                    BusinessObject busObj = (BusinessObject) stk.pop();
                    BusinessObject[] children = busObj.getBusObjChildren();
                    if (bo instanceof ObservationBlock) {
                        ObservationBlock ob = (ObservationBlock) bo;
                        FindingChart[] fcs = ob.getFindingCharts();
                        String srcFolderPath =
                            Persistence.getInstance().getRootFolderPath(
                                ob.getObsRun());
                        // copy finding charts to the root Folder 
                        // for an observing run.

                        ObservingRun destObsRun = folder.findObsRun();
                        if (ob.getObsRun().getId() != destObsRun.getId()) {
                            // TODO blank out inst comment here if
                            // run is not configured for it.
                            Instrument inst =
                                InstrumentList.getInstance().getInstrument(
                                    destObsRun.getInstCode(),
                                    destObsRun.getIPVersion());
                            if (!inst.hasInstComments()) {
                                ob.setInstrumentComments("");
                            }

                            for (int i = 0; i < fcs.length; i++) {
                                byte[] bdata =
                                    fcs[i].getFcDataLocal(srcFolderPath);
                                File f =
                                    new File(
                                        destFolderPath
                                            + File.separator
                                            + fcs[i].getCheckSum()
                                            + "."
                                            + fcs[i].getFcType());
                                if (!f.exists())
                                    FileIOUtils.writeBinaryFile(bdata, f);
                            }
                        }
                    }
                    for (int i = 0; i < children.length; i++) {
                        stk.push(children[i]);
                    }
                    objMgr.moveBusObj(
                        device_,
                        busObj.getId(),
                        busObj.getClass(),
                        newParent_);

                    bo =
                        (BusinessObject) objMgr.getBusObj(
                            device_,
                            busObj.getId(),
                            busObj.getClass());
                    bo.accept(new MoveUpdater(), folder.findObsRun());
                }

            } catch (ObjectNotFoundException e) {
                ErrorMessages.announceNoObject(baseFrame_, e);
            } catch (ObjectIOException e) {
                ErrorMessages.announceIOError(baseFrame_, e);
            } catch (IOException e) {
                ErrorMessages.announceIOError(baseFrame_, e);
            }
        }

        /**
         * Set the ObservingRun for any OBs or CBs that we move in case
         * they are being moved to a different observing run.
         */
        public class MoveUpdater extends VisitorAdaptor {

            public Object visit(CalibrationBlock cb, Object arg) {
                cb.setObsRun((ObservingRun) arg);
                return null;
            }

            public Object visit(ObservationBlock ob, Object arg) {
                ob.setObsRun((ObservingRun) arg);
                return null;
            }
        }
    }
}