package org.eso.ohs.phase2.apps.p2ppAppServer;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.Enumeration;
import java.util.Hashtable;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectManager;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;

/**
 * The abstract parent class of all storage managers. Defines operations that
 * the Persistence can expect of storage managers for all media.
 * 
 * <p>
 * $Id: RmiLocalManager.java,v 1.18 2004/12/14 17:47:35 tcanavan Exp $<br>
 * 
 * @author Bob Kemp
 * 
 * <br>
 * ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP
 * AppServer. <br>
 * ChangeLog: cmavros 20JUN02 1652/P2PP Reword error msg to include proxy
 * settings.
 * @see ObjectManager
 */

public class RmiLocalManager implements StorageManager {

    private static Logger stdlog_ = Logger.getLogger(RmiLocalManager.class);

    private static final String ERROR_MSG_TEXT =
        "The remote server is busy or unavailable\n"
            + "Your data could not be transferred. This is usually because\n"
            + "the server at ESO cannot be contacted, for some reason.\n"
            + "\n"
            + "Check your Internet connection to ESO and try again.\n\n"
            + "Note: If your firewall has limited communication ports it is \n"
            + "      possible to configure your http proxy in P2PP's site.cf file. \n"
            + "      For further information, refer to the P2PP Users' Manual.\n\n"
            + "If error persists, contact ESO User Support.";

    private static final int MAX_RETRIES = 10;

    private String serverURL_;

    private Media device_;

    private AppServerFactory factory_ = null;

    private RmiManagerInterface rmiMgr_;

    private DirectoryNode root_;

    private NodeMapper mapper_ = new NodeMapper();

    private int retryCount = 0;

    /** UserId of the currently logged-in user */
    private int userId_ = -1;

    private StringWriter sw_ = new StringWriter();

    public RmiLocalManager(Media device, String serverBaseURL)
        throws ObjectIOException {
        serverURL_ = serverBaseURL;
        device_ = device;

        stdlog_.debug(">>> RmiLocalManager: server URL is: " + serverURL_);
    }

    /**
     * Sets the id of the current user.
     * 
     * @param userId
     *            the id of the end-user
     */
    public void setUserId(int userId) {
        if (userId != userId_) {
            userId_ = userId;
            rmiMgr_ = null;
            root_ = null;
        }
    }

    /**
     * Gets the id of the current user.
     * 
     * @return the id of the end-user
     */
    public int getUserId() {
        return userId_;
    }

    /**
     * Get the root node of the DirectoryNode tree representing the directories
     * found on this medium. For example, a representation of the root of the
     * local StorableObject storage subtree.
     * 
     * @return a representation of the root node
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the method's execution.
     */
    public DirectoryNode getRoot() throws ObjectIOException {
        if (root_ == null) {
            checkConnection();
        }
        return root_;
    }

    /**
     * Verify that the TemplateSignature files used by the StorageManager are
     * the same version as those specified (and used by the client)
     * 
     * @param tsList
     *            array of version info objects for each template used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while checking
     */
    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tsList)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();
                String[] errs = rmiMgr_.verifyTemplateSignatures(tsList);
                logCompletedTransaction("verifyTemplateSignatures");
                return errs;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Verify that the TemplateSignature files used by the StorageManager are
     * the same version as those specified (and used by the client)
     * 
     * @param tsList
     *            array of version info objects for each template used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while checking
     */
    public FindingChart[] populateFindingCharts(long boId)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                FindingChart[] fcs = rmiMgr_.populateFindingCharts(boId);
                logCompletedTransaction("populateFindingCharts");
                return fcs;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Get the type of media maintained by this StorageManager instance.
     * 
     * @return an instance of the Media class.
     */
    public Media getMedia() {
        return device_;
    }

    /**
     * Calculates a unique id for a StorableObject.
     * 
     * @return Unique id
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the method's execution.
     */
    public long getNewId(Class mo, long baseId) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                long id = rmiMgr_.getNewId(mo, baseId);
                logCompletedTransaction("getNewId");
                return id;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Initialise the data structures for the appropiate media. For local
     * storage this will be a directory structure.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the method's execution.
     */
    public void initialise() throws ObjectIOException {
    }

    public boolean loginUser(String password) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                boolean loggedIn = rmiMgr_.loginUser(password);
                logCompletedTransaction("loginUser");

                return loggedIn;
            } catch (RuntimeException e) {
                e.printStackTrace(new PrintWriter(sw_));
                stdlog_.error(sw_.toString());
                return false;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Returns a DirectoryNode representing the path within the storage medium
     * of an object with the given id and type. For media that do not support
     * directories, the root directory may be returned.
     * 
     * @param id
     *            the id of the StorableObject.
     * @param objType
     *            the type of stored StorableObject.
     * 
     * @return Directory in which to find the object.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the method's execution.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     */
    public DirectoryNode find(long id, Class objType)
        throws ObjectNotFoundException, ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                DirectoryNode node = rmiMgr_.find(id, objType);
                DirectoryNode localNode = mapper_.getLocalNode(node);
                logCompletedTransaction("find");
                return localNode;
            } catch (RemoteException e) {
                resetServerConnection(e);
            }
        }
    }

    /**
     * Returns an array of objects representing the contents of an object with
     * the given id and type from the chosen directory. If the medium does not
     * support directories, the <code>node</code> parameter is ignored.
     * 
     * <p>
     * The array of objects contains:
     * <ul>
     * <li>as the first (0th) element, an instance of a StorableObject
     * subclass; followed by
     * <li>an alternating sequence of values:
     * <ul>
     * <li>at even indexes, a property name, ie a string X such that setX() and
     * getX() methods exist.
     * <li>at odd indexes, either an instance of java.lang.Long containing the
     * the id of the StorableObject subclass to which the property should refer
     * or a reference to the subclass itself.
     * </ul>
     * </ul>
     * 
     * @param node
     *            directory from which the object should be loaded.
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @return the array of objects and object references, as described above.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @exception ObjectIOException
     *                if an I/O error occurred while reading the object
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject read(
        DirectoryNode node,
        long id,
        Class objType)
        throws ObjectNotFoundException, ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                ObjectManager objMgr = ObjectManager.getObjectManager();
                UnpackedStorableObject objs = rmiMgr_.read(node, id, objType);

                logCompletedTransaction("read");
                return objs;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectNotFoundException e) {
                e.fillInStackTrace();
                throw e;
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Writes an array of objects with the given id and type to this storage
     * medium, in the chosen directory. If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node
     *            the directory in which the object should be saved.
     * @param id
     *            the id under which to store the StorableObject.
     * @param objType
     *            the type of StorableObject.
     * @param obj
     *            the array of objects and object references to store, as
     *            described in the section for the read method.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #update
     */
    public void write(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.write(node, id, objType, obj);
                logCompletedTransaction("write");
                return;
            } catch (RemoteException e) {
                if (e.detail instanceof RuntimeException) {
                    throw (RuntimeException) e.detail;
                }
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Updates an array of objects with the given id and type on this storage
     * medium, in the chosen directory. If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node
     *            the directory in which the object should be saved.
     * @param id
     *            the id under which to store the StorableObject.
     * @param objType
     *            the type of StorableObject.
     * @param obj
     *            the array of objects and object references to store, as
     *            described in the section for the read method.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #write
     */
    public void update(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.update(node, id, objType, obj);
                logCompletedTransaction("update");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Deletes an object with the given directory, id and type from this storage
     * medium. If the medium does not support directories, the <code>node</code>
     * parameter is ignored.
     * 
     * @param node
     *            directory from which the object should be deleted.
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #update
     */
    public void delete(DirectoryNode node, long id, Class objType)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.delete(node, id, objType);
                logCompletedTransaction("delete");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Deletes an object with the given directory, id and type from this storage
     * medium. If the medium does not support directories, the <code>node</code>
     * parameter is ignored.
     * 
     * @param node
     *            directory from which the object should be deleted.
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #update
     */
    public void move(
        long id,
        Class objType,
        DirectoryNode fromNode,
        DirectoryNode toNode)
        throws ObjectNotFoundException, ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.move(id, objType, fromNode, toNode);
                logCompletedTransaction("move");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            }
        }
    }

    /**
     * Create a directory in local store.
     * 
     * @param node
     *            The directory to be created
     * @param name
     *            The name of the directory to be created.
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *                if the method call generated an I/O error
     */
    public boolean createDirectory(DirectoryNode node, long id)
        throws ObjectIOException {

        for (;;) {
            try {
                checkConnection();

                boolean created = rmiMgr_.createDirectory(node, id);
                logCompletedTransaction("createDirectory");

                return created;
            } catch (RemoteException e) {
                resetServerConnection(e);
            }
        }
    }

    /**
     * Move a directory.
     * 
     * @param fromNode
     *            directory to be moved
     * @param toNode
     *            the new parent directory
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *                if the method call generated an I/O error
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     */
    public boolean moveDirectory(DirectoryNode fromNode, DirectoryNode toNode)
        throws ObjectIOException, ObjectNotFoundException {

        for (;;) {
            try {
                checkConnection();

                boolean moved = rmiMgr_.moveDirectory(fromNode, toNode);
                logCompletedTransaction("moveDirectory");

                return moved;
            } catch (RemoteException e) {
                resetServerConnection(e);
            }
        }
    }

    /**
     * Delete a directory from local storeage
     * 
     * @param node
     *            Delete a directory from Local Storeage
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *                if the method call generated an I/O error
     */
    public boolean deleteDirectory(DirectoryNode node)
        throws ObjectIOException {

        for (;;) {
            try {
                checkConnection();

                boolean deleted = rmiMgr_.deleteDirectory(node);
                logCompletedTransaction("deleteDirectory");
                return deleted;
            } catch (RemoteException e) {
                resetServerConnection(e);
            }
        }
    }

    /**
     * Returns a list of Summary objects associated with a directory.
     * 
     * @param node
     *            Directory to be used to build up list of objects
     * @return An array of ObjectFileInfo[] objects
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     */
    public Summary[] listObjects(DirectoryNode node) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                Summary[] summ = rmiMgr_.listObjects(node);
                logCompletedTransaction("listObjects");
                return summ;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Check in the specified objects in a single atomic operation.
     * 
     * @param usos
     *            Descriptions of each StorableObject to be checked in.
     * 
     * @return array of error strings. If length = 0, no errors.
     */
    public void checkIn(MultiObjSpec[] moi) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.checkIn(moi);
                logCompletedTransaction("checkIn");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Checkin the readme file for a given observing run
     * 
     * @param Readme
     *            file business object
     */

    public void checkInReadme(ObservingRun obsRun) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.checkInReadme(obsRun);
                logCompletedTransaction("checkInReadme");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }

    }

    /**
     * Checkout the readme file for a given observing run.
     * 
     * @param boId
     *            Observing run definition.
     */

    public String checkOutReadme(ObservingRun obsRun)
        throws ObjectIOException {
        String readme = null;
        try {
            checkConnection();

            readme = rmiMgr_.checkOutReadme(obsRun);
            logCompletedTransaction("checkOutReadme");
        } catch (RemoteException e) {
            resetServerConnection(e);
        } catch (ObjectIOException e) {
            e.fillInStackTrace();
            throw e;
        }
        return readme;
    }

    /**
     * Do a EVM verification test for the specified objects in a single atomic
     * operation.
     * 
     * @param usos
     *            Descriptions of each StorableObject to do EVM verification
     * 
     * @return boolean if OB passed EVM test
     */
    public boolean verifyObject(MultiObjSpec[] moi) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();
                boolean result = rmiMgr_.verifyObject(moi);
                logCompletedTransaction("verifyObject");
                return result;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * @return the requested IP as a byte array, or an empty (size 0) array if
     *         the IP does not exist on the server.
     * @exception ObjectIOException
     *                when there is an I/O exception.
     */
    public byte[] downloadInstrumentPackage(String instrument, float version)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                byte[] rc =
                    rmiMgr_.downloadInstrumentPackage(instrument, version);
                logCompletedTransaction("downloadInstrumentPackage");
                return rc;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Check out the ObsBlk or CalBlk with the specified id plus any attached
     * objects, ie Target, ObsDescr, etc.
     * 
     * @param id
     *            id of the ObsBlk or CalBlk to be checked out
     * 
     * @return CheckOutStatus object that contains a return status and (if
     *         successful) references to the checked-out objects.
     */
    public void checkOut(IdType[] idTypes)
        throws ObjectNotFoundException, ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.checkOut(idTypes);
                logCompletedTransaction("checkOut");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Write several object to the external media in an atomic operation.
     * 
     * @param moi
     *            Array of object descriptions.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #write
     * @see #update
     */
    public void multiStore(StorageManager.MultiObjSpec[] mos)
        throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.multiStore(mos);
                logCompletedTransaction("multiStore");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }
    }

    /**
     * Checks that we have a remote reference to the RmiStorageManager. If not,
     * then attempts are made to create a connection. If these attempts fail, an
     * ObjectIOException is thrown.
     * 
     * @exception ObjectIOException
     *                if the connection cannot be (re)created.
     */
    private void checkConnection() throws ObjectIOException {
        if (rmiMgr_ == null) {
            try {
                resetServerConnection(null);

                if (root_ == null) {
                    root_ = rmiMgr_.getRoot();

                    mapper_.rebuild(root_);
                }
            } catch (RemoteException e) {
                throw new ObjectIOException(e);
            }
        }
    }

    /**
     * Make repreated attempts to get a remote reference to an
     * RmiStorageManager. If these attempts fail, an ObjectIOException is
     * thrown.
     * 
     * @exception ObjectIOException
     *                if the connection cannot be (re)created.
     */
    private void resetServerConnection(RemoteException e)
        throws ObjectIOException {

        if (++retryCount > MAX_RETRIES) {
            retryCount = 0;
            e.printStackTrace(new PrintWriter(sw_));
            stdlog_.error(sw_.toString());
            throw new ObjectIOException(e, ERROR_MSG_TEXT);
        }

        stdlog_.debug(">>> resetServerConnection(): " + serverURL_);
        try {
            if (factory_ == null) {
                factory_ = (AppServerFactory) Naming.lookup(serverURL_);
            }
            try {
                // First try to get a new StorageManager instance.
                rmiMgr_ = factory_.getStorageManager(device_);
            } catch (RemoteException exc) {
                stdlog_.debug("Exception is caught");
                // Lost contact with server? Try reconnecting.
                factory_ = (AppServerFactory) Naming.lookup(serverURL_);
                rmiMgr_ = factory_.getStorageManager(device_);
            }
            stdlog_.debug("Setting user id " + getUserId());
            if (rmiMgr_ == null)
                stdlog_.debug("RMI manager = null");
            rmiMgr_.setUserId(getUserId());
        } catch (UnmarshalException exc) {
            throw new ObjectIOException(exc, ERROR_MSG_TEXT);
        } catch (NotBoundException exc) {
            exc.printStackTrace(new PrintWriter(sw_));
            stdlog_.error(sw_.toString());
            throw new ObjectIOException(exc, ERROR_MSG_TEXT);
        } catch (MalformedURLException exc) {
            exc.printStackTrace(new PrintWriter(sw_));
            stdlog_.error(sw_.toString());
            throw new ObjectIOException(exc, ERROR_MSG_TEXT);
        } catch (java.rmi.ConnectException exc) {
            exc.printStackTrace(new PrintWriter(sw_));
            stdlog_.error(sw_.toString());
            throw new ObjectIOException(exc, ERROR_MSG_TEXT);
        } catch (RemoteException exc) {
            exc.printStackTrace(new PrintWriter(sw_));
            stdlog_.error(sw_.toString());
            throw new ObjectIOException(exc, ERROR_MSG_TEXT);
        }

        stdlog_.debug("Reset server connection: done");
    }

    private void logCompletedTransaction(String caller) {
        stdlog_.debug("Completed call to " + caller);
    }

    private static class NodeMapper {
        private Hashtable pathToNode_ = new Hashtable();

        public DirectoryNode getLocalNode(DirectoryNode node) {
            return (DirectoryNode) pathToNode_.get(new Long(node.getId()));
        }

        public void rebuild(DirectoryNode root) {
            pathToNode_.clear();
            Enumeration enum = root.nodes();

            while (enum.hasMoreElements()) {
                DirectoryNode node = (DirectoryNode) enum.nextElement();
                pathToNode_.put(new Long(node.getId()), node);
            }
        }

    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#usdSubmit(org.eso.ohs.dfs.ObservingRun)
     */
    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();

                rmiMgr_.usdSubmit(obsRun);
                logCompletedTransaction("usdSubmit");
                return;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }

    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFile(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException {
        for (;;) {
            try {
                checkConnection();
                Readme readmefile = rmiMgr_.populateReadmeFile(obsRun);
                logCompletedTransaction("populateReadmeFile");
                return readmefile;
            } catch (RemoteException e) {
                resetServerConnection(e);
            } catch (ObjectIOException e) {
                e.fillInStackTrace();
                throw e;
            }
        }

    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#setReadmeFileStatus(java.lang.String, long)
     */
    public void setReadmeFileStatus(String status, long runId) throws ObjectIOException {
        throw new ObjectIOException ("Not available via the appserver"); 
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFileReplicate(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFileReplicate(ObservingRun obsRun) throws ObjectIOException {
        throw new ObjectIOException ("Not available via the appserver"); 
    }
}
