package org.eso.ohs.phase2.visibility;

import java.util.Date;
import java.util.Hashtable;


import com.mhuss.AstroLib.AstroDate;

/**
 * Class VisibilityCache
 * 
 * Caches visibility information.
 * Assumes constant observatory coordinates. The cache keys don't take into
 * account change of latitude and longitude.
 * 
 * @version $Revision: 1.3 $
 * @author ddorigo, Jul 8, 2005
 */

// $Id: VisibilityCache.java,v 1.3 2005/09/15 12:29:49 pnunes Exp $
public class VisibilityCache {

    /*
     * Cache of values
     */
    protected Hashtable moonRA_;

    protected Hashtable moonDEC_;

    protected Hashtable sunElevations_;

    protected Hashtable moonElevations_;

    protected Hashtable localSTimes_;

    /*
     * Cache of PCFs
     */
    protected Hashtable moonDownPCFs_;

    protected Hashtable sunElevationsPCFs_;

    protected Hashtable fliPCFs_;

    protected Hashtable timeIntervalPCFs_;

    protected Hashtable siderealTimeIntervalPCFs_;

    
    
    private static VisibilityCache visibilityCache_;

    /**
     * Initializes the class
     */
    protected VisibilityCache() {
        moonRA_ = new Hashtable();
        moonDEC_ = new Hashtable();
        sunElevations_ = new Hashtable();
        moonElevations_ = new Hashtable();
        localSTimes_ = new Hashtable();

        moonDownPCFs_ = new Hashtable();
        sunElevationsPCFs_ = new Hashtable();
        fliPCFs_ = new Hashtable();
        timeIntervalPCFs_ = new Hashtable();
        siderealTimeIntervalPCFs_ = new Hashtable();
    }

    public static VisibilityCache getInstance() {
        if (visibilityCache_ == null)
            visibilityCache_ = new VisibilityCache();

        return (visibilityCache_);

    }

    /**
     * Clears the chache 
     */
    public void clear() {
        moonRA_.clear();
        moonDEC_.clear();
        sunElevations_.clear();
        moonElevations_.clear();
        localSTimes_.clear();

        moonDownPCFs_.clear();
        sunElevationsPCFs_.clear();
        fliPCFs_.clear();
        timeIntervalPCFs_.clear();
        siderealTimeIntervalPCFs_.clear();
    }

    /**
     * Returns LST
     * 
     * @param jdate
     * @return Double or null if not cached
     */
    public Double getLST(double jdate) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double sidTime = (Double) localSTimes_.get(date);
        return sidTime;
    }

    /**
     * Returns sun elevation
     * 
     * @param jdate
     * @return Double or null if not cached
     */
    public Double getSunElevation(double jdate) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double sunElev = (Double) sunElevations_.get(date);
        return sunElev;
    }

    /**
     * Returns moon elevation 
     * @param jdate
     * @return Double or null if not cached
     */
    public Double getMoonElevation(double jdate) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonElev = (Double) moonElevations_.get(date);
        return moonElev;
    }

    /**
     * Returns moon RA
     * 
     * @param jdate
     * @return Double  or null if not cached
     */
    public Double getMoonRa(double jdate) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonRa = (Double) moonRA_.get(date);
        return moonRa;
    }

    /**
     * Returns moon DEC
     * @param jdate
     * @return Double or null if not cached
     */
    public Double getMoonDec(double jdate) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonDec = (Double) moonDEC_.get(date);
        return moonDec;
    }

    public void putLST(double jdate, double lst) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double lstObj = new Double(lst);
        localSTimes_.put(date, lstObj);
    }

    public void putSunElevation(double jdate, double sunEl) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double sunElObj = new Double(sunEl);
        sunElevations_.put(date, sunElObj);
    }

    public void putMoonElevation(double jdate, double moonEl) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonElObj = new Double(moonEl);
        moonElevations_.put(date, moonElObj);
    }

    public void putMoonRa(double jdate, double moonRa) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonRaObj = new Double(moonRa);
        moonRA_.put(date, moonRaObj);
    }

    public void putMoonDec(double jdate, double moonDec) {
        AstroDateExt date = new AstroDateExt(jdate);
        Double moonDecObj = new Double(moonDec);
        moonDEC_.put(date, moonDecObj);
    }

    public PCF getSunElevationPCF(Date start, Date end, int interval) {
        DateInterval date = new DateInterval(start, end, interval);
        PCF sunElPCF = (PCF) sunElevationsPCFs_.get(date);
        return sunElPCF;
    }

    public PCF getMoonDownPCF(Date start, Date end, int interval) {
        DateInterval date = new DateInterval(start, end, interval);
        PCF moonDownPCF = (PCF) moonDownPCFs_.get(date);
        return moonDownPCF;
    }

    public PCF getFLIPCF(Date start, Date end, int interval) {
        DateInterval date = new DateInterval(start, end, interval);
        PCF fliPCF = (PCF) fliPCFs_.get(date);
        return fliPCF;
    }

    public PCF getTimeIntervalPCF(Date start, Date end, int interval) {
        DateInterval date = new DateInterval(start, end, interval);
        PCF timePCF = (PCF) timeIntervalPCFs_.get(date);
        return timePCF;
    }

    public PCF getSiderealTimeIntervalPCF(Date start, Date end, int interval) {
        DateInterval date = new DateInterval(start, end, interval);
        PCF sidTimePCF = (PCF) siderealTimeIntervalPCFs_.get(date);
        return sidTimePCF;
    }

    public void putSunElevationPCF(Date start, Date end, int interval, PCF pcf) {
        DateInterval date = new DateInterval(start, end, interval);
        sunElevationsPCFs_.put(date, pcf);
    }

    public void putMoonDownPCF(Date start, Date end, int interval, PCF pcf) {
        DateInterval date = new DateInterval(start, end, interval);
        moonDownPCFs_.put(date, pcf);
    }

    public void putFLIPCF(Date start, Date end, int interval, PCF pcf) {
        DateInterval date = new DateInterval(start, end, interval);
        fliPCFs_.put(date, pcf);
    }

    public void putTimeIntervalPCF(Date start, Date end, int interval, PCF pcf) {
        DateInterval date = new DateInterval(start, end, interval);
        timeIntervalPCFs_.put(date, pcf);
    }

    public void putSiderealTimeIntervalPCF(Date start, Date end, int interval,
            PCF pcf) {
        DateInterval date = new DateInterval(start, end, interval);
        siderealTimeIntervalPCFs_.put(date, pcf);
    }

    /*
     * 
     * Class AstroDateExt
     * 
     * Represents AstroDateExt with equals
     * 
     * @version $Revision: 1.3 $ @author ddorigo, Jul 8, 2005
     */
    private class AstroDateExt extends AstroDate {
        /**
         * @param jdate
         */
        public AstroDateExt(double jdate) {         
           super(jdate);
        }

        public boolean equals(Object other) {
            boolean res = false;

            
            
            if (other instanceof AstroDateExt) {
            	AstroDateExt d = (AstroDateExt) other;
            	//res = toString().equals(d.toString());
            	//this.
            	res = ( Math.abs(jd()-d.jd()) < 0.0005 );
            /*
                AstroDateExt d = (AstroDateExt) other;
                res = (d.year() == this.year())
                        && (d.month() == this.month())
                        && (d.day() == this.day() )
                        
                        && (d.hour() == d.hour())
                        && (d.minute() == d.minute());
                        
                                
                 */               
            }

			return res;
 
        }

        
        public int hashCode() {
            return super.second();
        }
    }

    /*
     * 
     * Class DateInterval
     * 
     * Represents a pair of java.util.Date objects
     * 
     * @version $Revision: 1.3 $ @author ddorigo, Jul 8, 2005
     */
    private class DateInterval {

        private Date begin_;

        private Date end_;

        private int step_; // minutes

        /**
         * @param begin_
         * @param end_
         * @param step_
         */
        public DateInterval(Date begin_, Date end_, int step_) {
            this.begin_ = begin_;
            this.end_ = end_;
            this.step_ = step_;
        }

        public boolean equals(Object other) {
            boolean res = false;

            if (other instanceof DateInterval) {
                DateInterval d = (DateInterval) other;
                res = d.getBegin_().equals(begin_) && d.getEnd_().equals(end_)
                        && (d.getStep() == step_);
            }

            return (res);
        }

        public Date getBegin_() {
            return begin_;
        }

        public void setBegin_(Date begin_) {
            this.begin_ = begin_;
        }

        public Date getEnd_() {
            return end_;
        }

        public void setEnd_(Date end_) {
            this.end_ = end_;
        }

        /**
         * @param step
         *            The step to set.
         */
        public void setStep(int step) {
            this.step_ = step;
        }

        /**
         * @return Returns the step.
         */
        public int getStep() {
            return step_;
        }
        
        public int hashCode() {
            return (int) (step_+ begin_.getTime()+ end_.getTime());
        }
    }

}